/*
 *  $Header: C:\OOP\RCS\bt.cpv 1.6 1993/08/15 14:56:27 chris Exp chris $
 *
 *  Copyright (c) Chinchilla Software Ltd, 1991-1993. All rights reserved.
 *
 *  Wirth cloned Btree routines
 */

#include <stdio.h>
#include <iostream.h>
#include <strstrea.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include "bt.h"

// Search macro made a macro to for inlining
#define BTSEARCH(ep, l, m, r) { \
  l = 1; \
  r = c; \
  do { \
    int str_ret; \
    m = (l + r) >> 1; \
    str_ret = strcmp(ep->key, e[m - 1]->key); \
    if (str_ret <= 0) r = m - 1; \
    if (str_ret >= 0) l = m + 1; \
  } while (l <= r); \
}

BTEL *BTPAGE::insert(BTEL *ep)
{
  BTEL *retp;
  int i;
  BTPAGE  *np;
  int l, r;

  if (!this)
  {
    if (!(retp = new BTEL))
      return (BTEL *)0;
    retp->key = ep->key;
    retp->val = ep->val;
    retp->p = (BTPAGE *)0;

    return retp;
  }
  BTSEARCH(ep, l, i, r);
  if (l - r > 1)
    return (BTEL *)0;
  np = (!r) ? p : e[r - 1]->p;
  retp = np->insert(ep);
  if (retp)
  {
    ep = retp;
    r--;
    if (c < BTORDER)
    {
      for (i = c++ ; i != r + 1 ; i--)
	e[i] = e[i - 1];
      e[r + 1] = ep;

      return (BTEL *)0;
    }
    if (!(np = new BTPAGE))
      return (BTEL *)0;
    if (r < BTCHOP)
    {
      if (r == BTCHOP - 1)
	retp = ep;
      else
      {
	for (i = BTCHOP - 1, retp = e[i] ; i > r + 1 ; i--)
	  e[i] = e[i - 1];
	e[r + 1] = ep;
      }
      for (i = 0 ; i != BTCHOP ; i++)
	np->e[i] = e[i + BTCHOP];
    }
    else
    {
      for (r -= BTCHOP, retp = e[BTCHOP], i = 0 ; i != r ; i++)
	np->e[i] = e[i + BTCHOP + 1];
      for (np->e[r] = ep, i = r + 1 ; i != BTCHOP ; i++)
	np->e[i] = e[i + BTCHOP];
    }
    c = np->c = BTCHOP;
    np->p = retp->p;
    retp->p = np;
  }

  return retp;
}

int BTPAGE::underflow(BTPAGE *parent, int s)
{
  BTPAGE  *pp;
  int i;
  int parentc;
  int ppc;
  int k;

  if ((parentc = parent->c) > s + 1)
  {
    pp = parent->e[++s]->p;
    ppc = pp->c;
    e[BTCHOP - 1] = parent->e[s];
    e[BTCHOP - 1]->p = pp->p;
    if ((k = (ppc - BTCHOP + 1) >> 1) > 0)
    {
      /* Move items from pp to this */
      for (i = 0 ; i != k - 1 ; i++)
        e[i + BTCHOP] = pp->e[i];
      parent->e[s] = pp->e[k - 1];
      pp->p = pp->e[k - 1]->p;
      parent->e[s]->p = pp;
      for (i = 0, ppc -= k ; i != ppc ; i++)
        pp->e[i] = pp->e[i + k];
      pp->c = ppc;
      c = BTCHOP - 1 + k;

      return 0;
    }
    /* Merge pages this and pp */
    for (i = 0 ; i != BTCHOP ; i++)
      e[i + BTCHOP] = pp->e[i];
    for (i = s ; i != parentc - 1 ; i++)
      parent->e[i] = parent->e[i + 1];
    c = BTORDER;
    parent->c = parentc - 1;
    delete pp;

    return parent->c != BTCHOP;
  }
  /* pp is page to left of this */
  pp = (!s) ? parent->p : parent->e[s - 1]->p;
  ppc = pp->c + 1;
  if ((k = (ppc - BTCHOP) >> 1) > 0)
  {
    /* Move items from pp to this */
    for (i = BTCHOP - 2 ; i >= 0 ; i--)
      e[i + k] = e[i];
    e[k - 1] = parent->e[s];
    e[k - 1]->p = p;
    ppc -= k;
    for (i = k - 2 ; i >= 0 ; i--)
      e[i] = pp->e[i + ppc];
    p = pp->e[ppc - 1]->p;
    parent->e[s] = pp->e[ppc - 1];
    parent->e[s]->p = this;
    pp->c = ppc - 1;
    c = BTCHOP - 1 + k;

    return 0;
  }
  /* Merge pages this and pp (page to left of this) */
  pp->e[ppc - 1] = parent->e[s];
  pp->e[ppc - 1]->p = p;
  for (i = 0 ; i != BTCHOP - 1 ; i++)
    pp->e[i + ppc] = e[i];
  pp->c = BTORDER;
  parent->c = parentc - 1;
  delete this;

  return parent->c != BTCHOP;
}

int BTPAGE::del(BTPAGE *parent, int k)
{
  BTPAGE  *q;

  q = e[c - 1]->p;
  if (q)
    return (q->del(parent, k) ? q->underflow(this, c - 1) : 0);
  e[c - 1]->p = parent->e[k]->p;
  parent->e[k] = e[c - 1];

  return --c != BTCHOP;
}

int BTPAGE::delel(BTEL *ep)
{
  BTPAGE *child;
  int i;
  int l, r;

  if (!this)
    return 0;
  BTSEARCH(ep, l, i, r);
  child = (!r) ? p : e[r - 1]->p;
  if (l - r > 1)
  {
    l = i;
    if (!child)
    {
      delete e[i - 1]->key;
      e[i - 1]->key = (char *)0;
      delete e[i - 1];
      for (c--, i-- ; i != c ; i++)
        e[i] = e[i + 1];

      return c < BTCHOP;
    }
    return (child->del(this, l - 1) ? child->underflow(this, r - 1) : 0);
  }
  return child->delel(ep) ? child->underflow(this, r - 1) : 0;
}

BTEL *BTPAGE::next(BTEL *ep, int *last)
{
  BTPAGE *child;
  int m, l, r;
  BTEL  *retp;

  BTSEARCH(ep, l, m, r);
  child = (!r) ? p : e[r - 1]->p;
  if (l - r > 1)
  {
    *last = 0;
    if (!child)
    {
      if (m < c)
        return e[m];
      *last = 1;

      return e[m - 1];
    }
    return e[r]->p->first();
  }
  retp = child->next(ep, last);
  if (!(*last))
    return retp;
  if (r < c)
  {
    *last = 0;

    return(e[r]);
  }

  return e[r - 1];
}

BTEL *BTPAGE::prev(BTEL *ep, int *first)
{
  BTPAGE *child;
  int m, l, r;
  BTEL  *retp;

  BTSEARCH(ep, l, m, r);
  child = (!r) ? p : e[r - 1]->p;
  if (l - r > 1)
  {
    *first = 0;
    if (!child)
    {
      if (m > 1)
        return e[m - 2];
      *first = 1;

      return e[m - 1];
    }

    return child->last();
  }
  retp = child->next(ep, first);
  if (!(*first))
    return retp;
  if (r > 0)
  {
    *first = 0;

    return e[r - 1];
  }

  return e[r - 1];
}

BTEL *Btree::search(char *key)
{
  BTPAGE *rt;
  BTEL *ep;
  BTEL e;
  int i;
  int l, r, done;
  int str_ret;

  rt = pp;
  e.key = key;
  e.val = (void *)0;
  ep = &e;
  done = 0;
  do {
    if (rt)
    {
      l = 1;
      r = rt->c;
      do {
        i = (l + r) >> 1;
        str_ret = strcmp(ep->key, rt->e[i - 1]->key);
        if (str_ret <= 0) r = i - 1;
        if (str_ret >= 0) l = i + 1;
      } while (l <= r);
      if (l - r > 1)
      {
        ep = rt->e[i - 1];
	done = 1;
      }
      else
        rt = (!r) ? rt->p : rt->e[r - 1]->p;
    }
    else
      return (BTEL *)0;
  } while (!done);

  return ep;
}

BTEL *Btree::insert(char *key, void *val)
{
  BTPAGE *rt;
  BTPAGE  *old;
  BTEL  *ep;
  BTEL  e;

  rt = pp;
  if (!(e.key = new char[strlen(key) + 1]))
    return (BTEL *)0;
  strcpy(e.key, key);
  e.val = val;
  ep = rt->insert(&e);
  if (ep)
  {
    old = rt;
    if (!(rt = new BTPAGE))
      return (BTEL *)0;
    pp = rt;
    rt->c = 1;
    rt->p = old;
    rt->e[0] = ep;
  }
  return ep;
}

inline void Btree::drop(char *key, void *val)
{
  BTPAGE  *rt;
  BTPAGE  *old;
  BTEL  e;
  char save_str[256];

  rt = pp;
  strcpy(save_str, key);
  e.key = save_str;
  e.val = val;
  if (rt->delel(&e) && !rt->c)
  {
    old = rt;
    pp = old->p;
    delete old;
  }
}

inline BTEL *Btree::next(BTEL *ep)
{
  int last;
  BTEL  *retp;

  if (!pp)
    return (BTEL *)0;
  retp = pp->next(ep, &last);

  return (last) ? (BTEL *)0 : retp;
}

inline BTEL *Btree::prev(BTEL *ep)
{
  int first;
  BTEL  *retp;

  if (!pp)
    return (BTEL *)0;
  retp = pp->prev(ep, &first);

  return (first) ? (BTEL *)0 : retp;
}

void BTPAGE::pnode(int level)
{
  BTPAGE *pp;
  int i;

  pp = this;
  if (!pp)
    return;
  for (i = 0 ; i != level ; i++)
    cout << '\t';
  for (i = 0 ; i != pp->c ; i++)
    cout << pp->e[i]->key << " ";
  cout << endl;
  pp->p->pnode(level + 1);
  for (i = 0 ; i != pp->c ; i++)
    pp->e[i]->p->pnode(level + 1);
}

BTEL *BTPAGE::first()
{
  BTPAGE *p;

  if (!this)
    return (BTEL *)0;

  p = this;
  do {
    p = p->p;
  } while (p);

  return p->e[0];
}

BTEL *BTPAGE::last()
{
  BTPAGE *p;

  if (!this)
    return (BTEL *)0;

  p = this;
  do {
    p = p->e[p->c-1]->p;
  } while (p);

  return p->e[p->c-1];
}

/* Drop btree */
Btree::~Btree()
{
  BTEL  *ep;

  while ((ep = pp->first()) != (BTEL *)NULL)
    drop(ep->key, ep->val);
}

#ifdef MAIN

#define DEL_FREQ 10
#define LOOP_SZ  10000

main()
{
  int i;
  char ch[16], old_ch[DEL_FREQ][16];
  char *str;
  Btree bt;
  BTEL *el;

 for (i = 0; i != LOOP_SZ; i++)
 {
  if (i % 3)
    sprintf(ch, "%04dfred", i);
  else
    sprintf(ch, "fred%04d", i);
  el = bt.insert(ch, (void *)i);
  el = bt.search(ch);
  if (!(i % (LOOP_SZ / 10)) && i)
  {
    cout << "Key = " << el->get_key() <<
            ", el = " << (int)el->get_val() << endl;
  }
  if (i % DEL_FREQ)
  {
    if (i > DEL_FREQ)
      str = old_ch[i % DEL_FREQ];
    else
      str = ch;
    bt.drop(str, (void *)i);
    strcpy(old_ch[i % DEL_FREQ], ch);
  }
 }
  // bt.pnode(0);

  return 0;
}
#endif
