/*
 *	$Header: C:\OOP\RCS\bt.h_v 1.6 1993/08/15 14:56:30 chris Exp chris $
 *
 *	Copyright (c) Chinchilla Software Ltd, 1991-93. All rights reserved.
 *
 *	B-Tree routines.
 */
#ifndef BTREE_H
#define BTREE_H

/* The order of the Btree - an even number please */
#define BTORDPOWER 4
#define BTCHOP	(1 << BTORDPOWER)
#define BTORDER	(BTCHOP << 1)

/* The element structure describes an element in a Btree */
class BTEL
{
  private:

    char		*key;	/* Element's key */
    void		*val;	/* Element's value */
    class BTPAGE	*p;	/* Pointer to successor page */

  friend class BTPAGE;
  friend class Btree;

  public:

    char *get_key()
    {
      return key;
    }
    void *get_val()
    {
      return val;
    }
};

/* The page structure describes a page in a Btree */
class BTPAGE
{
  private:

    short int	c;		/* Number of elements in page */
    BTEL       	*e[BTORDER];	/* Array of elements */
    BTPAGE    	*p;		/* First successor page */

    BTEL *first();
    BTEL *last();
    BTEL *insert(BTEL *);
    BTEL *next(BTEL *, int *);
    BTEL *prev(BTEL *, int *);
    int del(BTPAGE *, int);
    int underflow(BTPAGE *, int);
    int delel(BTEL *);
    void pnode(int);

  friend class Btree;

  public:

};

/* The Btree itself */
class Btree
{
  private:

    BTPAGE *pp;

  public:

    Btree()
    {
      pp = (BTPAGE *)0;
    }
    ~Btree();

    BTEL *insert(char *, void *);
    BTEL *search(char *key);
    BTEL *next(BTEL *);
    BTEL *prev(BTEL *);
    void drop(char *, void *);
    void pnode(int i)
    {
      pp->pnode(i);
    }
    BTEL *first()
    {
      return pp->first();
    }
    BTEL *last()
    {
      return pp->last();
    }
};

#endif /* !BTREE_H */
