/*
 *  $Header: C:\PP2\BM\RCS\btree.c_v 1.1 1993/04/09 20:32:05 chris Exp chris $
 *
 *  Copyright (c) Chinchilla Software Ltd, 1991-1993. All rights reserved.
 *
 *  Wirth cloned Btree routines
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>

#include "btree.h"

/* Inline for speed */
#define BTSEARCH(ep, pp, l, m, r) { \
  l = 1; \
  r = pp->c; \
  do { \
    register int str_ret; \
    m = (l + r) >> 1; \
    str_ret = strcmp(ep->key, pp->e[m - 1]->key); \
    if (str_ret <= 0) r = m - 1; \
    if (str_ret >= 0) l = m + 1; \
  } while (l <= r); \
}

static BTEL *insert(BTEL *ep, BTPAGE *pp)
{
  int l;
  int r;
  int m;
  register BTEL  *retp;
  BTPAGE  *np;
  register int i;

  if (!pp)
  {
    if (!(retp = (BTEL *)malloc(sizeof(BTEL))))
      return (BTEL *)0;
    retp->key = ep->key;
    retp->val = ep->val;
    retp->p = (BTPAGE *)0;

    return retp;
  }
  BTSEARCH(ep, pp, l, m, r);
  if (l - r > 1)
    return (BTEL *)0;
  retp = insert(ep, (!r) ? pp->p : pp->e[r - 1]->p);
  if (retp)
  {
    ep = retp;
    r--;
    if (pp->c < BTORDER)
    {
      for (i = pp->c++ ; i != r + 1 ; i--)
	pp->e[i] = pp->e[i - 1];
      pp->e[r + 1] = ep;

      return (BTEL *)0;
    }
    if (!(np = (BTPAGE *)malloc(sizeof(BTPAGE))))
      return (BTEL *)0;
    if (r < BTCHOP)
    {
      if (r == BTCHOP - 1)
	retp = ep;
      else
      {
	for (i = BTCHOP - 1, retp = pp->e[i] ; i > r + 1 ; i--)
	  pp->e[i] = pp->e[i - 1];
	pp->e[r + 1] = ep;
      }
      for (i = 0 ; i != BTCHOP ; i++)
	np->e[i] = pp->e[i + BTCHOP];
    }
    else
    {
      for (r -= BTCHOP, retp = pp->e[BTCHOP], i = 0 ; i != r ; i++)
	np->e[i] = pp->e[i + BTCHOP + 1];
      for (np->e[r] = ep, i = r + 1 ; i != BTCHOP ; i++)
	np->e[i] = pp->e[i + BTCHOP];
    }
    pp->c = np->c = BTCHOP;
    np->p = retp->p;
    retp->p = np;
  }

  return retp;
}

static int underflow(BTPAGE *parent, BTPAGE **uflowp, int s)
{
  register BTPAGE  *pp;
  register BTPAGE  *uf;
  register int i;
  int parentc;
  int ppc;
  int k;

  uf = *uflowp;
  if ((parentc = parent->c) > s + 1)
  {
    pp = parent->e[++s]->p;
    ppc = pp->c;
    uf->e[BTCHOP - 1] = parent->e[s];
    uf->e[BTCHOP - 1]->p = pp->p;
    if ((k = (ppc - BTCHOP + 1) >> 1) > 0)
    {
      /* Move items from pp to uf */
      for (i = 0 ; i != k - 1 ; i++)
        uf->e[i + BTCHOP] = pp->e[i];
      parent->e[s] = pp->e[k - 1];
      pp->p = pp->e[k - 1]->p;
      parent->e[s]->p = pp;
      for (i = 0, ppc -= k ; i != ppc ; i++)
        pp->e[i] = pp->e[i + k];
      pp->c = ppc;
      uf->c = BTCHOP - 1 + k;

      return 0;
    }
    /* Merge pages uf and pp */
    for (i = 0 ; i != BTCHOP ; i++)
      uf->e[i + BTCHOP] = pp->e[i];
    for (i = s ; i != parentc - 1 ; i++)
      parent->e[i] = parent->e[i + 1];
    uf->c = BTORDER;
    parent->c = parentc - 1;
    (void)free((void *)pp);

    return parent->c != BTCHOP;
  }
  /* pp is page to left of uf */
  pp = (!s) ? parent->p : parent->e[s - 1]->p;
  ppc = pp->c + 1;
  if ((k = (ppc - BTCHOP) >> 1) > 0)
  {
    /* Move items from pp to uf */
    for (i = BTCHOP - 2 ; i >= 0 ; i--)
      uf->e[i + k] = uf->e[i];
    uf->e[k - 1] = parent->e[s];
    uf->e[k - 1]->p = uf->p;
    ppc -= k;
    for (i = k - 2 ; i >= 0 ; i--)
      uf->e[i] = pp->e[i + ppc];
    uf->p = pp->e[ppc - 1]->p;
    parent->e[s] = pp->e[ppc - 1];
    parent->e[s]->p = uf;
    pp->c = ppc - 1;
    uf->c = BTCHOP - 1 + k;

    return 0;
  }
  /* Merge pages uf and pp (page to left of uf) */
  pp->e[ppc - 1] = parent->e[s];
  pp->e[ppc - 1]->p = uf->p;
  for (i = 0 ; i != BTCHOP - 1 ; i++)
    pp->e[i + ppc] = uf->e[i];
  pp->c = BTORDER;
  parent->c = parentc - 1;
  (void)free((void *)uf);

  return parent->c != BTCHOP;
}

static int del(BTPAGE *pp, BTPAGE *parent, int k)
{
  BTPAGE  *q;

  q = pp->e[pp->c - 1]->p;
  if (q)
    return (del(q, parent, k) ? underflow(pp, &q, pp->c - 1) : 0);
  pp->e[pp->c - 1]->p = parent->e[k]->p;
  parent->e[k] = pp->e[pp->c - 1];

  return --pp->c != BTCHOP;
}

static int delel(BTEL *ep, BTPAGE *pp)
{
  register int i;
  BTPAGE  *daughter;
  int m;
  int l;
  int r;

  if (!pp)
    return 0;
  BTSEARCH(ep, pp, l, m, r);
  daughter = (!r) ? pp->p : pp->e[r - 1]->p;
  if (l - r > 1)
  {
    if (!daughter)
    {
      (void)free((void *)pp->e[m - 1]->key);
      pp->e[m - 1]->key = (char *)0;
      (void)free((void *)pp->e[m - 1]);
      for (pp->c--, i = m - 1 ; i != pp->c ; i++)
        pp->e[i] = pp->e[i + 1];

      return pp->c < BTCHOP;
    }
    return (del(daughter, pp, m - 1) ? underflow(pp, &daughter, r - 1) : 0);
  }
  return delel(ep, daughter) ? underflow(pp, &daughter, r - 1) : 0;
}

static BTEL *next(BTEL *ep, BTPAGE *pp, int *last)
{
  BTPAGE  *daughter;
  int m, l, r;
  BTEL  *retp;

  BTSEARCH(ep, pp, l, m, r);
  daughter = (!r) ? pp->p : pp->e[r - 1]->p;
  if (l - r > 1)
  {
    *last = 0;
    if (!daughter)
    {
      if (m < pp->c)
        return pp->e[m];
      *last = 1;

      return pp->e[m - 1];
    }
    return bt_first(pp->e[r]->p);
  }
  retp = next(ep, daughter, last);
  if (!(*last))
    return retp;
  if (r < pp->c)
  {
    *last = 0;

    return(pp->e[r]);
  }

  return pp->e[r - 1];
}

static BTEL *prev(BTEL *ep, BTPAGE *pp, int *first)
{
  BTPAGE  *daughter;
  int m, l, r;
  BTEL  *retp;

  BTSEARCH(ep, pp, l, m, r);
  daughter = (!r) ? pp->p : pp->e[r - 1]->p;
  if (l - r > 1)
  {
    *first = 0;
    if (!daughter)
    {
      if (m > 1)
        return pp->e[m - 2];
      *first = 1;

      return pp->e[m - 1];
    }

    return bt_last(daughter);
  }
  retp = next(ep, daughter, first);
  if (!(*first))
    return retp;
  if (r > 0)
  {
    *first = 0;

    return pp->e[r - 1];
  }

  return pp->e[r - 1];
}

void bt_pnode(BTPAGE *pp, int level)
{
  int i;

  if (!pp)
    return;
  for (i = 0 ; i != level ; i++)
    (void) putchar('\t');
  for (i = 0 ; i != pp->c ; i++)
    (void) printf("%s  ", pp->e[i]->key);
  (void) putchar('\n');
  bt_pnode(pp->p, level + 1);
  for (i = 0 ; i != pp->c ; i++)
    bt_pnode(pp->e[i]->p, level + 1);
}

BTEL *bt_search(char *key, BTPAGE *pp)
{
  BTEL  *ep;
  BTEL  e;
  int m;
  int l;
  int r;
  int done;

  e.key = key;
  e.val = (void *)0;
  ep = &e;
  done = 0;
  do {
    if (pp)
    {
      BTSEARCH(ep, pp, l, m, r);
      if (l - r > 1)
      {
        ep = pp->e[m - 1];
	done = 1;
      }
      else
        pp = (!r) ? pp->p : pp->e[r - 1]->p;
    }
    else
      return (ep = (BTEL *)0);
  } while (!done);

  return ep;
}

BTEL *bt_insert(char *key, void *val, BTPAGE **root)
{
  register BTPAGE *rt;
  BTPAGE  *old;
  BTEL  *ep;
  BTEL  e;

  rt = *root;
  if (!(e.key = (char *)malloc((unsigned)(strlen(key) + 1))))
    return (BTEL *)0;
  strcpy(e.key, key);
  e.val = val;
  ep = insert(&e, rt);
  if (ep)
  {
    old = rt;
    if (!(rt = (BTPAGE *)malloc(sizeof(BTPAGE))))
      return (BTEL *)0;
    *root = rt;
    rt->c = 1;
    rt->p = old;
    rt->e[0] = ep;
  }
  return ep;
}

void bt_delete(char *key, void *val, BTPAGE **root)
{
  register BTPAGE  *rt;
  BTPAGE  *old;
  BTEL  e;
  char save_str[256];

  rt = *root;
  strcpy(save_str, key);
  e.key = save_str;
  e.val = val;
  if (delel(&e, rt) && !rt->c)
  {
    old = rt;
    *root = old->p;
    (void)free((void *)old);
  }
}

BTEL *bt_first(BTPAGE *pp)
{
  if (!pp)
    return (BTEL *)0;

  return (pp->p) ? bt_first(pp->p) : pp->e[0];
}

BTEL *bt_last(BTPAGE *pp)
{
  if (!pp)
    return (BTEL *)0;

  return (pp->p) ? bt_last(pp->e[pp->c-1]->p) : pp->e[pp->c-1];
}

BTEL *bt_next(BTEL *ep, BTPAGE *pp)
{
  int last;
  BTEL  *retp;

  if (!pp)
    return (BTEL *)0;
  retp = next(ep, pp, &last);

  return (last) ? (BTEL *)0 : retp;
}

BTEL *bt_prev(BTEL *ep, BTPAGE *pp)
{
  int first;
  BTEL  *retp;

  if (!pp)
    return (BTEL *)0;
  retp = prev(ep, pp, &first);

  return (first) ? (BTEL *)0 : retp;
}

/* Drop btree */
void bt_fold(BTPAGE **pp)
{
  BTEL  *ep;

  while ((ep = bt_first(*pp)) != (BTEL *)NULL)
    bt_delete(ep->key, ep->val, pp);
}

#ifdef MAIN

#define DEL_FREQ 10

main()
{
  int i;
  char ch[16], old_ch[DEL_FREQ][16];
  char *str;
  BTPAGE *bt = (BTPAGE *)0;
  BTEL *el;

 for (i = 0; i != 10000; i++)
 {
  if (i % 3)
    sprintf(ch, "%04dfred", i);
  else
    sprintf(ch, "fred%04d", i);
  el = bt_insert(ch, (void *)i, &bt);
  el = bt_search(ch, bt);
/*
*/
  if (!(i % 1000) && i)
  {
    printf("Key = %s, el = %d\n", el->key, (int)el->val);
  }
  if (i % DEL_FREQ)
  {
    if (i > DEL_FREQ)
      str = old_ch[i % DEL_FREQ];
    else
      str = ch;
    bt_delete(str, (void *)i, &bt);
    strcpy(old_ch[i % DEL_FREQ], ch);
  }
 }
  // bt_pnode(bt, 0);
 /*
 */
  bt_fold(&bt);

  return 0;
}
#endif
