Program BtrieveErrors;

uses
  Btv, BtvConst,
  WinDos, WinTypes, WinProcs,
  OWindows, ODialogs, OStdDlgs,
  Strings, Validate;

{$R- X+ D+ L+}
{$R BTVERR.RES}
{$I BTVERR.INC}

const
  id_ErrorNumber = 101;
  id_ErrorName   = 102;
  id_ErrorText   = 103;
  id_SaveBtn     = 104;
  id_ClearBtn    = 105;
  id_NextBtn     = 106;
  id_PrevBtn     = 107;
  id_LoadBtn     = 108;


type
  PBtrvErrorRec = ^TBtrvErrorRec;
  TBtrvErrorRec = record
    ErrorNumber : Integer;
    ErrorName   : String[80];
    ErrorText   : Array[0..900] of Char;
  end;


  TBtrvApp = object(TApplication)
    procedure InitMainWindow; virtual;
  end;


  PMultiEdit	= ^TMultiEdit;
  TMultiEdit = object(TEdit)
    procedure WMKeyDown(var Msg: TMessage);
      virtual  wm_First + wm_KeyDown;
  end;


  PError  = ^TError;
  TError  = Object(ErrorDisplay)
    Function    Display(Error     : Integer;
                        ErrorMsg  : String;
                        OpCode    : Byte;
                        OpCodeMsg : String;
                        FileName  : PathStr
                        ): ErrorAction;             Virtual;
  end;


  PBtrvWindow = ^TBtrvWindow;
  TBtrvWindow = object(TWindow)
    ErrorNumber : PEdit;
    ErrorName   : PEdit;
    ErrorText   : PEdit;
    FileName    : array[0..fsPathName] of Char;
    FileOpen    : Boolean;
    F           : PBtrieveFile;
    ErrDisplay  : PError;
    ErrHandler  : PErrorHandler;
    Data        : PBtrvErrorRec;
    NewRecord   : Boolean;

    constructor Init(AParent: PWindowsObject; ATitle: PChar);
    destructor Done; virtual;
    function CanClose: Boolean; virtual;
    function HasChanged: Boolean;

    { button methods }
    procedure IDLoadBtn(var Msg : TMessage);
      virtual id_First + id_LoadBtn;
    procedure IDNextBtn(var Msg : TMessage);
      virtual id_First + id_NextBtn;
    procedure IDPrevBtn(var Msg : TMessage);
      virtual id_First + id_PrevBtn;
    procedure IDSaveBtn(var Msg : TMessage);
      virtual id_First + id_SaveBtn;
    procedure IDClearBtn(var Msg : TMessage);
      virtual id_First + id_ClearBtn;

    { menu methods }
    procedure CMFileNew(var Msg: TMessage);
      virtual cm_First + cm_FileNew;
    procedure CMFileOpen(var Msg: TMessage);
      virtual cm_First + cm_FileOpen;
    procedure CMFileClose(var Msg: TMessage);
      virtual cm_First + cm_FileClose;
  end;


var
  BtrvApp: TBtrvApp;


procedure TBtrvApp.InitMainWindow;
  begin
    MainWindow := New(PBtrvWindow, Init(nil, 'Btrieve Error Editor'));
  end;


function TError.Display(Error     : Integer;
                        ErrorMsg  : String;
                        OpCode    : Byte;
                        OpCodeMsg : String;
                        FileName  : PathStr
                        ): ErrorAction;

  var
    Temp : Array[0..512] of Char;
    Op   : String[10];
    Err  : String[10];
    Reply: Integer;

  begin
    Str(Error, Err);
    Str(OpCode, Op);
    StrPCopy(Temp, 'Btrieve Error # ' + Err + ' - ' + ErrorMsg + #13#10 +
                   'Btrieve Opcode # ' + Op + ' - ' + OpCodeMsg + #13#10);
    Reply := MessageBox(BtrvApp.MainWindow^.HWindow, Temp, 'BTRIEVE IO ERROR',
                        mb_AbortRetryIgnore or mb_IconQuestion);

    Case Reply of
      id_Abort : Display := erAbort;
      id_Retry : Display := erRetry;
      id_Ignore: Display := erDone;
    end;
  end;


procedure TMultiEdit.WMKeyDown(var Msg: TMessage);

  var
    nKeyState : Integer;

  begin
    if (Msg.WParam = vk_Tab) then
    begin
      nKeyState := GetAsyncKeyState(vk_Shift);

      if ($FF00 and nKeyState <> 0) then
        SetFocus(Previous^.HWindow)
      else
        SetFocus(Next^.HWindow);
    end

    else
      Inherited WmKeyDown(Msg);
  end;


constructor TBtrvWindow.Init(AParent: PWindowsObject; ATitle: PChar);

  var
    AStat: PStatic;
    ABtn : PButton;

  begin
    inherited Init(AParent, ATitle);
    Attr.Menu := LoadMenu(HInstance, MakeIntResource(Main_Menu));
    Attr.X    := 10;
    Attr.y    := 10;
    Attr.W    := 655;
    Attr.H    := 470;
    ErrorNumber := New(PEdit, Init(@Self, id_ErrorNumber, '0', 120, 10, 70, 24, 6, False));
    ErrorNumber^.Validator := New(PRangeValidator, Init(0,5000));
    ErrorName   := New(PEdit, Init(@Self, id_ErrorName, '', 120, 50, 500, 24, 81, False));
    ErrorText   := New(PMultiEdit, Init(@Self, id_ErrorText, '', 120, 90, 500, 240, 901, True));
    ABtn  := New(PButton, Init(@Self, id_SaveBtn, 'Save', 120, 365, 60, 30, False));
    ABtn  := New(PButton, Init(@Self, id_LoadBtn, 'Load', 210, 365, 60, 30, False));
    ABtn  := New(PButton, Init(@Self, id_ClearBtn,'Clear',300, 365, 60, 30, False));
    ABtn  := New(PButton, Init(@Self, id_NextBtn, 'Next', 390, 365, 60, 30, False));
    ABtn  := New(PButton, Init(@Self, id_PrevBtn, 'Prev', 480, 365, 60, 30, False));
    AStat := New(PStatic, Init(@Self, 200, 'Error Number :', 10, 10, 110, 24, 0));
    AStat := New(PStatic, Init(@Self, 201, 'Error Name :', 10, 50, 110, 24, 0));
    AStat := New(PStatic, Init(@Self, 202, 'Error Text :', 10, 90, 110, 24, 0));

    ErrorText^.Attr.Style := ErrorText^.Attr.Style and not ws_HScroll and not es_AutoHScroll;
    EnableKBHandler;
    FileOpen := False;
    FileName[0] := #0;
    F := nil;
    ErrDisplay := New(PError, Init);
    ErrHandler := New(PDefErrorHandler, Init(ErrDisplay));
    ErrHandler^.AddErrors([bEOF]);
    New(Data);
    NewRecord := True;
  end;

destructor TBtrvWindow.Done;
  begin
    if FileOpen then
    begin
      F^.Close;
      Dispose(F, Done);
    end;

    Dispose(ErrDisplay, Done);
    Dispose(ErrHandler, Done);
    inherited Done;
  end;

procedure TBtrvWindow.IDLoadBtn(var Msg : TMessage);

  var
    Temp : Array[0..80] of Char;
    Num  : Integer;
    Err  : Integer;

  begin
    if not FileOpen then
    begin
      MessageBox(HWindow, 'Please open a file first.', 'Information',
                 mb_OK or mb_IconExclamation);
      EXIT;
    end;

    ErrorNumber^.GetText(Temp, 6);
    Val(Temp, Num, Err);

    if (Err = 0) then
    begin
      F^.MakeKey(@Num, nil,nil,nil,nil,nil);
      F^.Get(bGetEqual, bNoLock);

      if (F^.bResult = bOkay) then
      begin
        Str(Data^.ErrorNumber, Temp);
        ErrorNumber^.SetText(Temp);
        ErrorName^.SetText(StrPCopy(Temp, Data^.ErrorName));
        ErrorText^.Transfer(@Data^.ErrorText, tf_SetData);
        NewRecord := False;
      end;
    end;
  end;

procedure TBtrvWindow.IDSaveBtn(var Msg : TMessage);

  var
    Temp : Array[0..1024] of Char;
    Num  : Integer;
    Err  : Integer;
    i,x  : Integer;
    Len  : Integer;
    dc : hdc;

  begin
    if not FileOpen then
    begin
      MessageBox(HWindow, 'Please open a file first.', 'Information',
                 mb_OK or mb_IconExclamation);
      EXIT;
    end;

    ErrorNumber^.GetText(Temp, 6);
    Val(Temp, Num, Err);

    if (Err <> 0) then
       EXIT

    else if NewRecord then
    begin
      F^.ErrorsOnOff(False);
      F^.MakeKey(@Num, nil,nil,nil,nil,nil);
      F^.Get(bGetEqual + bGetKey, bNoLock);
      F^.ErrorsOnOff(True);

      if (F^.bResult = bOkay) then
      begin
        Num := MessageBox(HWindow, 'Do you want to overwrite?', 'Record Already Exists',
                          mb_YesNo or mb_IconQuestion);

        if (Num = id_Yes) then
        begin
          NewRecord := False;
          F^.Get(bGetEqual, bNoLock);
        end

        else
          EXIT;
      end;
    end;

    F^.ClearBuffer;
    Data^.ErrorNumber := Num;
    ErrorName^.GetText(Temp, 81);
    Data^.ErrorName := StrPas(Temp);
    Len := ErrorText^.GetTextLen;
    ErrorText^.GetText(Temp, Len + 1);
    x := 0;

    for i := 0 to Len - 1 do
      Case Temp[i] of
        #13 :;
        #10 :
        begin
          Data^.ErrorText[x] := ' ';
          Inc(x);
        end

        else
        begin
          Data^.ErrorText[x] := Temp[i];
          Inc(x);
        end;
      end;
(*
    DC := GetDC(HWindow);
    TextOut(DC, 0,350, Data^.ErrorText, x);
    ReleaseDC(HWindow, DC);
*)

    if NewRecord then
      F^.Insert
    else
      F^.Update;

    ErrorNumber^.SetText('0');
    ErrorName^.SetText('');
    ErrorText^.SetText('');
    NewRecord := True;
    SetFocus(ErrorNumber^.HWindow);
    ErrorNumber^.SetSelection(0,1);
  end;

procedure TBtrvWindow.IDPrevBtn(var Msg : TMessage);

  var
    Temp : Array[0..80] of Char;

  begin
    if CanClose and FileOpen and not NewRecord then
    begin
      F^.Get(bGetPrev, bNoLock);

      if (F^.bResult = bOkay) then
      begin
        Str(Data^.ErrorNumber, Temp);
        ErrorNumber^.SetText(Temp);
        ErrorName^.SetText(StrPCopy(Temp, Data^.ErrorName));
        ErrorText^.SetText(Data^.ErrorText);
        NewRecord := False;
      end;
    end;
  end;

procedure TBtrvWindow.IDNextBtn(var Msg : TMessage);

  var
    Temp : Array[0..80] of Char;

  begin
    if CanClose and FileOpen and not NewRecord then
    begin
      F^.Get(bGetNext, bNoLock);

      if (F^.bResult = bOkay) then
      begin
        Str(Data^.ErrorNumber, Temp);
        ErrorNumber^.SetText(Temp);
        ErrorName^.SetText(StrPCopy(Temp, Data^.ErrorName));
        ErrorText^.SetText(Data^.ErrorText);
        NewRecord := False;
      end;
    end;
  end;

procedure TBtrvWindow.IDClearBtn(var Msg : TMessage);
  begin
    if not FileOpen or CanClose then
    begin
      ErrorNumber^.SetText('0');
      ErrorName^.SetText('');
      ErrorText^.SetText('');
      NewRecord := True;
      SetFocus(ErrorNumber^.HWindow);
      ErrorNumber^.SetSelection(0,1);
    end;
  end;

procedure TBtrvWindow.CMFileOpen(var Msg: TMessage);

  var
    FD : PFileDialog;

  begin
    if CanClose then
      if not FileOpen then
      begin
        FD := New(PFileDialog, Init(@Self, MakeIntResource(sd_FileOpen), FileName));
        StrCopy(FileName, '*.ERR');

        if Application^.ExecDialog(FD) = id_OK then
        begin
          F := New(PBtrieveFile, Init(StrPas(FileName), ErrHandler, Data, SizeOf(TBtrvErrorRec)));
          F^.Open(bNormal, '');
          FileOpen := (F^.bResult = bOkay);
        end
      end

      else
      begin
        MessageBox(HWindow, 'Please close the current file.', 'Information',
                   mb_OK or mb_IconExclamation);
      end;
  end;

function TBtrvWindow.CanClose: Boolean;

  var
    Reply: Integer;

  begin
    CanClose := True;

    if HasChanged then
    begin
      Reply := MessageBox(HWindow, 'Do you want to save?', 'Current Record has Changed',
                          mb_YesNo or mb_IconQuestion);

      if (Reply = id_Yes) then
         CanClose := False;
    end;
  end;

procedure TBtrvWindow.CMFileNew(var Msg: TMessage);

  var
    FD : PFileDialog;

  begin
    if FileOpen then
    begin
      MessageBox(HWindow, 'Please close the current file.', 'Information',
                 mb_OK or mb_IconExclamation);
      EXIT;
    end;

    FD := New(PFileDialog,Init(@Self, MakeIntResource(sd_FileOpen), FileName));
    FD^.Caption := 'File New';
    StrCopy(FileName, 'BTRIEVE.ERR');

    if (Application^.ExecDialog(FD) = id_OK) then
    begin
      F := New(PBtrieveFile, Init(StrPas(FileName), ErrHandler, Data, SizeOf(TBtrvErrorRec)));
      F^.AddKeySegment(1, SizeOf(Integer), bExtended, bInteger, 0, 0);
      F^.Create(bNormal, SizeOf(TBtrvErrorRec), 1024, 0, bNoOverWrite);

      if (F^.bResult = bOkay) then
        F^.Open(bNormal, '');

      FileOpen := (F^.bResult = bOkay);
    end;
  end;

procedure TBtrvWindow.CMFileClose(var Msg: TMessage);
  begin
    F^.Close;
    FileOpen := not (F^.bResult = bOkay);

    if not FileOpen then
      Dispose(F, Done);
  end;

function TBtrvWindow.HasChanged: Boolean;
  begin
    HasChanged := ErrorNumber^.IsModified or ErrorName^.IsModified or ErrorText^.IsModified;
  end;


begin
  BtrvApp.Init('BTRVERR');
  BtrvApp.Run;
  BtrvApp.Done;
end.