/******************************************************************************
*                                                                             *
*    FILE: CString2.CPP                                                       *
*                                                                             *
*    PURPOSE: CString2 Implementation File.  CString2 is an extension to the  *
*             MFC's CString. It provides some stripping capabilities and      *
*             individual character type queries                               *
*                                                                             *
*******************************************************************************
*                                       *                                     *
* Author:     Robert J. Trembley        *  Language: MSVC++ v1.0, MFC 2.0     *
* Internet:   bob.trembley@angus.mi.org *                                     *
* Compuserve: 71726,766                 *  CString2 Version 0.90              *
*                                       *                                     *
******************************************************************************/

#include "CString2.H"

/******************************************************************************
*                                                                             *
*    FUNCTION:   CString2::StripPrecedingCharacters                           *
*                                                                             *
*    PARAMETERS: cChar   - character to be stripped                           *
*    RETURNS:    Nothing                                                      *
*    PURPOSE:    Strips all occurances of the passed character from the       *
*                 beginning of the string                                     *
*                                                                             *
******************************************************************************/

void CString2::StripPrecedingCharacters (char cChar)
{
   ASSERT(cChar != '\0');

   if (!m_nDataLength)
      {
      TRACE0 ("CString2::StripPrecedingCharacters : NULL String Passed.\n");
      return;
      }

   char* szTemp = (char*) new char[m_nDataLength];

   for (;;) // ever
      {
      if (!m_nDataLength)
         break;

      if (m_pchData[0] == cChar)
         {
         _fstrcpy (szTemp, m_pchData+1);
         _fstrcpy (m_pchData, szTemp);

         --m_nDataLength;
         }
      else
         break;
      }

   if (szTemp)
      delete szTemp;

return;
}

/******************************************************************************
*                                                                             *
*    FUNCTION:   CString2::StripTrailingCharacters                            *
*                                                                             *
*    PARAMETERS: cChar   - character to be stripped                           *
*    RETURNS:    Nothing                                                      *
*    PURPOSE:    Strips all occurances of the passed character from the       *
*                 end of the string                                           *
*                                                                             *
******************************************************************************/

void CString2::StripTrailingCharacters (char cChar)
{
   ASSERT(cChar != '\0');

   if (!m_nDataLength)
      {
      TRACE0 ("CString2::StripTrailingCharacters : NULL String Passed.\n");
      return;
      }

   for (;;) //ever
     {
      if (!m_nDataLength)
         break;

     if (m_pchData[m_nDataLength-1] == cChar)
        m_pchData[--m_nDataLength] = '\0';
     else
        break;
     }

return;
}

/******************************************************************************
*                                                                             *
*    FUNCTION:   CString2::ExpandTabs                                         *
*                                                                             *
*    PARAMETERS: nTabSize - number of blanks to insert for each tab           *
*                 default == 8                                                *
*    RETURNS:    Nothing                                                      *
*    PURPOSE:    Replaces all tab characters with the number of spaces        *
*                 specified by nTabSize                                       *
*                                                                             *
******************************************************************************/

void CString2::ExpandTabs (int nTabSize /*=8*/)
{
   ASSERT(nTabSize > 0);

   for (int nIndex = 0; nIndex < m_nDataLength; nIndex++)
      {
      if (IsTab(nIndex))
         {
         if (nIndex)
            {
            //--------------------------------
               CString szTemp;
               CString szTab (' ', nTabSize);
            //--------------------------------

            szTemp = this->Left(nIndex);                       // copy up to the tab
            szTemp += szTab;                                   // replace the tab
            szTemp += this->Right(m_nDataLength - nIndex - 1); // copy the rest
            this->operator = ((const char*) szTemp);           // Replace it
            nIndex += nTabSize - 1;
            }
         }
      }
}
