/******************************************************************************
*                                                                             *
*    FILE: CString2.H                                                         *
*                                                                             *
*    PURPOSE: CString2 Interface File.  CString2 is an extension to the       *
*             MFC's CString. It provides some stripping capabilities and      *
*             individual character type queries                               *
*                                                                             *
*******************************************************************************
*                                       *                                     *
* Author:     Robert J. Trembley        *  Language: MSVC++ v1.0, MFC 2.0     *
* Internet:   bob.trembley@angus.mi.org *                                     *
* Compuserve: 71726,766                 *  CString2 Version 0.90              *
*                                       *                                     *
******************************************************************************/

#ifndef _CSTRING_2_
#define _CSTRING_2_

#ifndef __AFX_H__
#include <Afx.H>
#endif

#ifndef _INC_CTYPE
#include <Ctype.H>
#endif

#define _TAB_CHAR 0x09

/******************************************************************************
*                                                                             *
*    CLASS: CString2                                                          *
*                                                                             *
******************************************************************************/

class CString2 : public CString
{
public:

   // Constructors (need to pass on parameters to base class)
   //
   CString2 ();
   CString2 (const CString& stringSrc);
   CString2 (char ch, int nRepeat = 1);
   CString2 (const char* psz);
   CString2 (const char* pch, int nLength);

#ifdef _NEARDATA
   // Additional versions for far string data
   //
   CString2 (LPCSTR lpsz);
   CString2 (LPCSTR lpch, int nLength);
#endif

   // Character Query Member Functions
   //
   int IsTab          (int nIndex) const;
   int IsWhiteSpace   (int nIndex) const;
   int IsAlphabetic   (int nIndex) const;
   int IsAlphanumeric (int nIndex) const;
   int IsBlank        (int nIndex) const;
   int IsPunctuation  (int nIndex) const;
   int IsUppercase    (int nIndex) const;
   int IsLowercase    (int nIndex) const;
   int IsNumeric      (int nIndex) const;
   int IsHexNumeric   (int nIndex) const;


   // Misc Functions
   //
   void MakeCharUpper (int nIndex);
   void MakeCharLower (int nIndex);


   // Tab Expansion
   //
   void ExpandTabs (int nTabSize = 8);


   // Stripping Functions
   //
   void StripTrailingBlanks  ();
   void StripPrecedingBlanks ();

   void StripTrailingCharacters  (char cChar);
   void StripPrecedingCharacters (char cChar);
};

/******************************************************************************
*                                                                             *
*    CString2 inline functions                                                *
*                                                                             *
******************************************************************************/

//===== Constructors =====
//
inline CString2::CString2() : CString() {}
inline CString2::CString2(const CString& stringSrc) : CString(stringSrc) {}
inline CString2::CString2(char ch, int nRepeat) : CString(ch, nRepeat) {}
inline CString2::CString2(const char* psz) : CString(psz) {}
inline CString2::CString2(const char* pch, int nLength) : CString(pch, nLength) {}

#ifdef _NEARDATA
inline CString2::CString2(LPCSTR lpsz) : CString(lpsz) {}
inline CString2::CString2(LPCSTR lpch, int nLength) : CString(lpch, nLength) {}
#endif


//===== Character Query Member Functions =====
//
inline int CString2::IsTab (int nIndex) const // Non Zero if ==, 0 if !=
{
   ASSERT(nIndex >= 0);
   ASSERT(nIndex < m_nDataLength);

   return m_pchData[nIndex] == _TAB_CHAR;
}

inline int CString2::IsAlphabetic (int nIndex) const // Non Zero if ==, 0 if !=
{
   ASSERT(nIndex >= 0);
   ASSERT(nIndex < m_nDataLength);

   return isalpha(m_pchData[nIndex]);
}

inline int CString2::IsAlphanumeric (int nIndex) const // Non Zero if ==, 0 if !=
{
   ASSERT(nIndex >= 0);
   ASSERT(nIndex < m_nDataLength);

   return isalnum(m_pchData[nIndex]);
}

inline int CString2::IsWhiteSpace (int nIndex) const // Non Zero if ==, 0 if !=
{
   ASSERT(nIndex >= 0);
   ASSERT(nIndex < m_nDataLength);

   return isspace(m_pchData[nIndex]);
}

inline int CString2::IsBlank (int nIndex) const // Non Zero if ==, 0 if !=
{
   ASSERT(nIndex >= 0);
   ASSERT(nIndex < m_nDataLength);

   return m_pchData[nIndex] == ' ';
}

inline int CString2::IsPunctuation (int nIndex) const // Non Zero if ==, 0 if !=
{
   ASSERT(nIndex >= 0);
   ASSERT(nIndex < m_nDataLength);

   return ispunct(m_pchData[nIndex]);
}

inline int CString2::IsUppercase (int nIndex) const // Non Zero if ==, 0 if !=
{
   ASSERT(nIndex >= 0);
   ASSERT(nIndex < m_nDataLength);

   return isupper(m_pchData[nIndex]);
}

inline int CString2::IsLowercase (int nIndex) const // Non Zero if ==, 0 if !=
{
   ASSERT(nIndex >= 0);
   ASSERT(nIndex < m_nDataLength);

   return islower(m_pchData[nIndex]);
}

inline int CString2::IsNumeric (int nIndex) const // Non Zero if ==, 0 if !=
{
   ASSERT(nIndex >= 0);
   ASSERT(nIndex < m_nDataLength);

   return isdigit(m_pchData[nIndex]);
}

inline int CString2::IsHexNumeric (int nIndex) const // Non Zero if ==, 0 if !=
{
   ASSERT(nIndex >= 0);
   ASSERT(nIndex < m_nDataLength);

   return isxdigit(m_pchData[nIndex]);
}



//===== Misc Functions =====
//
inline void CString2::MakeCharUpper (int nIndex)
{
   ASSERT(nIndex >= 0);
   ASSERT(nIndex < m_nDataLength);

   m_pchData[nIndex] = toupper(m_pchData[nIndex]);
}

inline void CString2::MakeCharLower (int nIndex)
{
   ASSERT(nIndex >= 0);
   ASSERT(nIndex < m_nDataLength);

   m_pchData[nIndex] = tolower(m_pchData[nIndex]);
}



//===== Stripping Functions =====
//
inline void CString2::StripPrecedingBlanks ()
   { StripPrecedingCharacters (' '); }


inline void CString2::StripTrailingBlanks ()
   { StripTrailingCharacters (' '); }


#endif  //_CSTRING_2_
