{$A+,B-,D+,E+,F-,G+,I+,L+,N+,P-,Q+,R+,S+,T-,V+,X+,Y+}
{$M 16384,0}
{**************************************************************************}
{*                             INFINT.PAS                                 *}
{**************************************************************************}
{*                                                                        *}
{*   Object-Oriented toolkit for dealing with large integers.             *}
{*   (only limited by Heap-Space)                                         *}
{*                                                                        *}
{*                                                                        *}
{*                                                                        *}
{*                                                                        *}
{*------------------------------------------------------------------------*}
{*   Autoren          : Christian Sperr                                   *}
{*   Datum            : 01.07.1993 (based on Infint.BIB)                  *}
{*   Letzte Aenderung :                                                   *}
{**************************************************************************}
UNIT InfInt;

INTERFACE

  USES STRINGS, OBJECTS;

  TYPE
    InfIntT       = ^InfIntZiffer;
    InfIntZiffer = RECORD
                      Z    : Longint;   {integer} {CS. 1.7.93}
                      next : InfIntT;
                   END;
    PInfInt  = ^TInfInt;
    TInfInt  = OBJECT (TObject)
                 Number : InfIntT;
                 CONSTRUCTOR Init;
                 CONSTRUCTOR AssignStr(s:PChar);
                 CONSTRUCTOR Assign(X:PInfInt);
                 CONSTRUCTOR AssignLong(X:Longint);
                 CONSTRUCTOR AssignComp(X:comp);
                 CONSTRUCTOR Sum(X,Y:PInfInt);
                 CONSTRUCTOR Diff(X,Y:PInfInt);
                 CONSTRUCTOR Prod(X,Y:PInfint);
                 CONSTRUCTOR Quot(X,Y:PInfInt);
                 CONSTRUCTOR Remainder(X,Y:PInfInt);
                 CONSTRUCTOR GGT(X,Y:PInfInt);
                 DESTRUCTOR  Done; VIRTUAL;
                 PROCEDURE SetZero;
                 FUNCTION Sign:integer;
                 PROCEDURE ChangeSign;
                 PROCEDURE AssignNew(X:PinfInt);
                 PROCEDURE Add(X:PInfInt);
                 PROCEDURE Sub(X:PInfInt);
                 PROCEDURE Multiply(X:PInfInt);
                 FUNCTION DivideRemainder(Denom:PinfInt): PInfInt;
                 PROCEDURE MultiplyLong(X:longInt);
                 FUNCTION DivideRemainderLong(X:longint):longint;
                 PROCEDURE ShiftLeft(n:longint);
                 PROCEDURE ShiftRight(n:longint);
                 PROCEDURE Power(exp:longint);
                 PROCEDURE Sqrt;
                 FUNCTION Compare(X:PInfInt):integer;
                 FUNCTION Str:PChar;
                 PROCEDURE Write(VAR F:Text);
                 CONSTRUCTOR Load(VAR S:TStream);
                 PROCEDURE Store(VAR S:TStream);
               END;


  CONST
    RInfInt : TStreamRec = (
                 ObjType: 106;
                 VmtLink: Ofs(TypeOf(TInfInt)^);
                 Load   : @TInfInt.Load;
                 Store  : @TInfInt.Store
                 );


  PROCEDURE RegisterInfInt;

IMPLEMENTATION


CONST Basis : longint = 100000000;

{*****************************************************************************}
{*                         Allgemeine Prozedur                               *}
{*****************************************************************************}

FUNCTION Sign(A:longint):integer;
BEGIN
  IF A < 0 THEN Sign := -1 ELSE Sign := 1;
END;

{*****************************************************************************}
{*                        Operationen mit InfIntT's                           *}
{*****************************************************************************}

PROCEDURE IIInvert(VAR x : InfIntT);
{ kehrt die Reihenfolge der Ziffern um                                      }
  VAR p,q,r : InfIntT;
BEGIN
  IF x <> nil THEN
  BEGIN
    r := nil; p := x^.next;q := x;
    WHILE q <> nil DO
    BEGIN
      q^.next := r;
      r := q;
      q := p;
      IF q <> nil THEN p := q^.next;
    END;
    x := r;
  END
END; {IIInvert}

PROCEDURE IISetZero(VAR x : InfIntT);
{Setzt x = 0 dh der Speicherplatz wird freigegeben                          }
  VAR  q : InfIntT;
BEGIN
  IF x <> nil THEN
  BEGIN
    q := x^.next;
    WHILE x <> nil DO
    BEGIN
      dispose(x);
      x := q;
      IF q <> NIL THEN q := q^.next;
    END;
    x := nil;
  END
END;{IISetZero}

FUNCTION IISign(X:InfIntT):integer;
BEGIN
  IF X = nil THEN IISign := 0
  ELSE
  BEGIN
    WHILE (x^.z = 0) AND (x^.next <> nil) DO x := x^.next;
    IF x^.z  < 0 THEN IISign := -1 ELSE IISign := 1;
  END;
END;{IISign}

{.PA}
FUNCTION IICopy(x : InfIntT):InfIntT;
{liefert eine Kopie von x                                                   }
  VAR p,q : InfIntT;
BEGIN
  p := nil;
  WHILE x <> nil DO
  BEGIN
    new(q);
    q^.z := x^.z;
    q^.next := p;
    p := q;
    x := x^.next;
  END;
  IIInvert(p);
  IICopy := p;
END;{IICopy}

PROCEDURE IINeg(VAR x:InfIntT);
{Multipliziert x mit -1                                                     }
  VAR p : InfIntT;
BEGIN
  p := x;
  WHILE p <> nil DO
  BEGIN
    p^.z := -p^.z;
    p := p^.next;
  END;
END;{IINeg}
{.A-}

FUNCTION IISum(A,B:InfIntT):InfIntT;
{Liefert die Summe von A und B ohne diese zu verndern                      }
  VAR Ap,Bp,Cp,Cp2,LastNotZero : InfIntT;
      positiv                  : boolean;
      u                        : longint;

  PROCEDURE Uebertrag;
  { Berechnet den bertrag und das Vorzeichen des Ergebnisses               }
  BEGIN
    u := 0;
    IF Cp^.z > (Basis - 1) THEN
    BEGIN
      Cp^.z := Cp^.z - Basis;
      u := 1;
    END;
    IF Cp^.z < -(Basis-1) THEN
    BEGIN
      Cp^.z := Cp^.z + Basis;
      u := -1;
    END;
    IF (Cp^.z < 0) THEN positiv := false
    ELSE IF Cp^.z > 0 THEN positiv := true; {Vorzeichen des Ergebnisses     }
  END;{Uebertrag}

{.PA}
BEGIN
  {Beide Zahlen addieren und Vorzeichen des Ergebnisses feststellen         }
  Ap := A; Bp := B; positiv := true;
  IF A = nil THEN Cp := IICopy(B)
  ELSE IF B = nil THEN Cp := IICopy(A);
  new(Cp); Cp2 := Cp; Cp^.next := nil;
  u := 0;
  WHILE (Ap <> nil) AND (Bp <> nil) DO
  BEGIN
    Cp^.z := Ap^.z +{-} Bp^.z + u;   {- Fr Differenz von A und B           }
    Uebertrag;
    new(Cp^.next); Cp := Cp^.next; Cp^.next := nil;
    Ap := Ap^.next; Bp := Bp^.next;
  END;
  IF Ap = nil THEN Ap := Bp; {Rest des Ergebnisses kopieren                  }
  WHILE Ap <> nil DO
  BEGIN
    Cp^.z := Ap^.z + u;
    Uebertrag;
    new(Cp^.next); Cp := Cp^.next; Cp^.next := nil;
    Ap := Ap^.next;
  END;
  Cp^.z := u;
  {Jetzt sind A und B addiert, aber nicht alle Ziffern haben das richtige   }
  {Vorzeichen. Darum wird das Ergebnis noch normalisiert                    }
  u := 0; Cp := Cp2; LastNotZero := Cp2;
  IF positiv THEN
    WHILE Cp <> nil DO
    BEGIN
      Cp^.z := Cp^.z + u;
      u := 0;
      IF Cp^.z < 0 THEN
      BEGIN
        Cp^.z := Cp^.z + Basis;
        u := -1;
      END;
      IF Cp^.z <> 0 THEN LastNotZero := Cp;
      Cp := Cp^.next;
    END
  ELSE
    WHILE Cp <> nil DO
    BEGIN
      Cp^.z := Cp^.z + u;
      u := 0;
      IF Cp^.z > 0 THEN
      BEGIN
        Cp^.z := Cp^.z - Basis;
        u := 1;
      END;
      IF Cp^.z <> 0 THEN LastNotZero := Cp;
      Cp := Cp^.next;
    END;
  {berflssige Nullen Ende Lschen                                         }
  IISetZero(LastNotZero^.next);
  IF (Cp2^.z = 0) AND (Cp2^.next = nil) THEN
  BEGIN
    IISum := nil;
    dispose(Cp2);
  END
  ELSE
  IISum := Cp2;
END; {IISum }

FUNCTION IIDiff(A,B:InfIntT):InfIntT;
{Berechnet die Differenz zweier Zahlen                                      }
BEGIN
  IF A = B THEN
  BEGIN
    IIDiff := nil;
    exit;
  END;
  IINeg(B);
  IIDiff := IISum(A,B);
  IINeg(B);
END;{IIDiff}
{Schneller wre eigene Prozedur wie IISum                               }

PROCEDURE IntProd(x,y:longint;VAR HiWord, LoWord : longint);
{Berechnet das doppelt lange Produkt von X und Y mittels real Arithmetik  }
{Sollte unter Umstnden durch ein Inline Statement ersetzt werden         }
{24.03.1993 verwenden von Longints anstatt Reals. }
{01.07.1993 muliplikation von 2 comp-Zahlen nach (a + zb) * (c + zd) = bdzz + (a+c)z +ac}

{ VAR C : longint;
BEGIN
  C :=  longint(X) * longint(Y);
  HiWord := C DIV Basis;
  LoWord := C MOD Basis;
END;}{IntProd}

  VAR
    c,cx,cy : comp;
  BEGIN
    cx := x; cy := y;
    c := cx * cy;
    HiWord := TRUNC(c / basis);
    cy := HiWord;
    LoWord := trunc(c - (cy * Basis));
    IF abs(LoWord) > Basis THEN Halt(254);
  END;
{.PA}
FUNCTION IISProd(A:InfIntT;B:longint):InfIntT;
VAR Ap,Cp,Cl,erg:InfIntT;
    u,u1,vb,va : longint;
BEGIN
  va := IISign(A);
  CASE va OF
    -1 : IINeg(A);
     0 : BEGIN
           IISProd := nil;
           exit
         END;
  END;
  IF B = 0 THEN
  BEGIN
    IF va < 0 THEN IINeg(A);
    IISProd := nil;
    exit;
  END;
  vb := Sign(B);
  IF vb < 0 THEN B := -B;
  new(Cp); erg := Cp; Cl := nil;
  Ap := A; u := 0;
  WHILE Ap <> nil DO
  BEGIN
    WITH Cp^ DO
    BEGIN
      IntProd(Ap^.z,B,u1,z);
      z := z + u;
      u := z DIV Basis + u1;
      z := z MOD Basis;
      new(next);
    END;
    Cl := Cp;
    Cp := Cp^.next;
    Ap := Ap^.next;
  END;
  WHILE u > 0 DO
  BEGIN
    Cp^.z := u MOD Basis;
    u := u DIV Basis;
    Cl := Cp;
    new(Cp^.next);
    Cp := Cp^.next;
  END;
  Cl^.next := nil;
  dispose(Cp);
  IF va <> vb THEN IINeg(erg);
  IF va = -1 THEN IINeg(A);
  IISProd := erg;
END;{IISProd}

FUNCTION IIProd(A,B : InfIntT):InfIntT;
{Berechnet A * B                                                            }
  VAR Ap,Bp,Cp,Cp2,Erg : InfIntT;
      H,L,k,u,va,vb    : longint;

BEGIN
  IF (A = nil) OR (B = nil) THEN IIProd := nil
  ELSE
  BEGIN
    Ap := A; WHILE Ap^.z = 0 DO Ap := Ap^.next;
    IF Ap^.z < 0 THEN
    BEGIN
     va := -1;
     IINeg(A);
    END
    ELSE va := 1;
    IF A <> B THEN
    BEGIN
      Bp := B;
      WHILE Bp^.z = 0 DO Bp := Bp^.next;
      IF Bp^.z < 0 THEN
      BEGIN
       vb := -1;
       IINeg(B);
      END
      ELSE vb := 1;
    END ELSE vb := va;
    Ap := A;
    new(Cp2); Cp := Cp2; Erg := Cp;
    Cp^.next := nil; Cp^.z := 0;
    {Hier beginnt der eigentliche Multiplikations-Algorithmus               }
    WHILE Ap <> nil DO
    BEGIN
      Bp := B; u := 0; Cp := Cp2;
      WHILE Bp <> nil DO {innere Schleife                                   }
      BEGIN
        IntProd(Ap^.z,Bp^.z,H,L);
        k := L + u + Cp^.z;
        Cp^.z := k MOD Basis;
        u := k DIV Basis + H;
        IF Cp^.next = nil THEN
        BEGIN
          new(Cp^.next);
          Cp^.next^.next := nil;
          Cp^.next^.z := 0;
        END;
        Cp := Cp^.next;
        Bp := Bp^.next;
      END; {Ende der inneren Schleife                                       }
      Cp^.z := u;
      Cp2 := Cp2^.next;
      Ap := Ap^.next;
    END; {Ende der usseren Schleife                                        }
    IF u = 0 THEN {dann ist die hchste Ziffer 0 und wird gelscht          }
    BEGIN
      Cp := Erg; Cp2 := Erg;
      WHILE Cp2^.next <> nil DO
      BEGIN
        Cp := Cp2;
        Cp2 := Cp2^.next;
      END;
      Cp^.next := nil;
      dispose(Cp2);
    END;
    {Jetzt mssen die Vorzeichen wieder in Ordnung gebracht werden          }
    IF va = -1 THEN IINeg(A);
    IF (vb = -1) AND (A <> B) THEN IINeg(B);
    IF va <> vb THEN IINeg(Erg);
    IIProd := Erg;
  END; {ELSE}
END; {IIProd}
 
PROCEDURE IISDiv(A:InfIntT;B:longint;VAR Q:InfIntT;VAR R:longint);
VAR va,vb : longint;
    Ap,Qp : InfIntT;
    D     : comp;
BEGIN
  va := IISign(A);
  vb := Sign(B);
  IF b = 0 THEN
  BEGIN
    writeln(#7,'Fehler: Division durch 0');
    Q := nil;
    R := -1;
    exit;
  END;
  IF va = 0 THEN
  BEGIN
    Q := nil;
    R := 0;
    exit;
  END;
  IF va < 0 THEN IINeg(A);
  B := abs(B);
  IIInvert(A);
  Ap := A; new(Q); Qp := Q;
  D := 0;
  REPEAT
    D := D * Basis + Ap^.z;
    Ap := Ap^.next;
  UNTIL (D >= B) OR (Ap = nil);
  IF D < B THEN
  BEGIN {Divisor grsser als Dividend}
    IIInvert(A);
    Q := nil;
    dispose(Qp);
    IF va < 0 THEN IINeg(A);
    R := round(D);
    exit
  END;
  WHILE Ap <> nil DO
  BEGIN
    Qp^.z := trunc(D/B);
    D := D - 1.0 * Qp^.z * B;
    new(Qp^.next);
    Qp := Qp^.next;
    D := D * Basis + Ap^.z;
    Ap := Ap^.next;
  END;
  Qp^.z := trunc(D/B);
  D := D - 1.0 * Qp^.z * B;
  Qp^.next := nil;
  R := trunc(D);
  IIInvert(Q);
  IIInvert(A);
  IF va < 0 THEN
  BEGIN
    IINeg(A);
    R := -R;
  END;
  IF va <> vb THEN IINeg(Q);
END;{IISDiv}
 
PROCEDURE IIDiv(A,B:InfIntT;VAR Q,R:InfIntT);
VAR Rp,Bp,BN,RP2,Qp :InfIntT;
    mr : comp;
    u,d,b1,b2,qd,va,vb : longint;
BEGIN
  IF B = nil THEN
  BEGIN
    writeln(#7,'Fehler: Division durch 0');
    Q := nil; R := nil;
    exit;
  END;
  IF A = nil THEN
  BEGIN
    Q := nil;
    R := nil;
    exit;
  END;
  IF B^.next = nil THEN
  BEGIN {Lnge von B = 1 -> einfacher Algorithmus}
    new(R);
    IISDiv(A,B^.z,Q,R^.z);
    R^.next := nil;
    IF R^.z = 0 THEN
    BEGIN
      dispose(R);
      R := nil;
    END;
    exit;
  END;
  vb := IISign(B);
  IF vb < 0 THEN IINeg(B);
  R := IICopy(A);
  va := IISign(A);
  IF va < 0 THEN IINeg(R);
  {hier beginnt der eigentliche Divisions-Algorithmus}
  {1. Normalisierung}
  IIInvert(B);
  d := Basis DIV (B^.z +1);
  IIInvert(B);
  IF d <> 1 THEN
  BEGIN
    BN := IISProd(B,d);
    Rp := IISProd(R,d);
    IISetZero(R); R := Rp;
  END
  ELSE BN := B;
  IIInvert(R);IIInvert(BN);
  IF BN^.z <= R^.z THEN
  BEGIN
    RP := R;
    new(R);
    R^.z := 0;
    R^.next := RP;
  END;
  {Ende der Normalisierung,Dividend R und Divisor BN sind invertiert }
  RP2 := R;
  BP := BN;
  REPEAT RP2 := RP2^.next; Bp := Bp^.next UNTIL (Bp = nil) OR (RP2 = nil);
  IF Rp2 = nil THEN
  BEGIN {In diesem Fall ist A < B}
    IF d > 1 THEN IISetZero(BN) ELSE IIInvert(BN);
    IF R^.z = 0 THEN
    BEGIN
      Rp := R^.next;
      dispose(R);
      R := Rp;
    END;
    IIInvert(R);
    IF d > 1 THEN
    BEGIN
      IISDiv(R,d,Rp,d);
      IISetZero(R);
      R := Rp;
    END;
    Q := nil;
    IF vb < 0 THEN IINeg(B);
    IF va < 0 THEN IINeg(R);
    exit;
  END;
  b1 := BN^.z;b2 := Bn^.next^.z; IIInvert(BN);
  {BN wieder in der normalen Reihenfolge }
  new(Q); Qp := Q;
  WHILE RP2 <> nil DO
  BEGIN              {Hauptschleife der Division}
    {Quotientenziffer schtzen}
    mr := 1.0 * Basis * R^.z + R^.next^.z; {die beiden hchsten Ziffern}
    IF R^.z >= B1 THEN qd := Basis - 1
    ELSE qd := trunc(mr/B1);
    {Schneller Test ob qd zu gross. Erkennt nicht alle Flle}
    WHILE 1.0 * b2 * qd > (mr - 1.0 * qd * b1) * Basis + R^.next^.next^.z DO qd := pred(qd);
    IIInvert(R); {R auch wieder in normaler Reihenfolge}
    {Zurckmultiplizieren und gleichzeitig Subtrahieren}
    Rp := RP2; { Rp2 zeigt auf die 'tiefste' Stelle, die verndert wird}
    Bp := Bn;
    u := 0;
    WHILE Bp <> nil DO
    BEGIN
      mr := 1.0 * qd * Bp^.z +u;
      u := trunc(mr/Basis);
      Rp^.z := Rp^.z - trunc(mr - 1.0 * Basis * u);
      IF Rp^.z < 0 THEN
      BEGIN
        Rp^.z := Rp^.z + Basis;
        u := u + 1;
      END;
      Bp := Bp^.next;
      Rp := Rp^.next;
    END;
    Rp^.z := Rp^.z - u;
    IF Rp^.z < 0 THEN
    BEGIN
      Rp^.z := Rp^.z + Basis;
      u := 1;
    END
    ELSE u := 0;
    {Falls hier u > 0 ist, so ist das Ergebnis hier negativ also qd um
     1 zu gross. Tritt dies ein wird qd um eins erniedrigt und BN zu R
     zurckaddiert}
    IF u > 0 THEN
    BEGIN
      qd := pred(qd);
      u := 0;
      Rp := Rp2;
      Bp := BN;
      WHILE Bp <> nil DO
      BEGIN
        Rp^.z := Rp^.z + Bp^.z +u;
        IF Rp^.z >= Basis THEN
        BEGIN
          Rp^.z := Rp^.z - Basis;
          u := 1
        END
        ELSE u := 0;
        Bp := Bp^.next;
        Rp := Rp^.next;
      END;
      Rp^.z := (Rp^.z + u) MOD Basis;
    END;
    {der hier auftretende Uebertrag hebt sich mit dem auf der oben nicht
     subtrahiert wurde}
    {R muss jetzt um die hchste Stelle, die jetzt null ist gekrzt werden}
    Rp := R;
    WHILE Rp^.next^.next <> nil DO Rp := Rp^.next;
    dispose(Rp^.next);
    Rp^.next := nil;
    Qp^.z := qd; new(Qp^.next); Qp := Qp^.next;
    IIInvert(R); {R wieder in umgekehrter Reihenfolge}
    Rp2 := Rp2^.next;
  END; {WHILE Rp2 <> nil}
  WHILE (R <> nil) AND (R^.z = 0) DO
  BEGIN
    Rp := R;
    R := R^.next;
    dispose(Rp);
  END;
  IIInvert(R);
  Qp^.next := nil;
  IIInvert(Q);
  Qp := Q; Q := Q^.next; dispose(Qp);
  {jetzt mssen die Vorzeichen wieder in Ordnung gebracht werden}
  IF d > 1 THEN
  BEGIN
    IISDiv(R,d,Rp,d);
    IISetZero(R);
    R := Rp;
  END;
  IF va < 0 THEN IINeg(R);
  IF vb < 0 THEN IINeg(B);
  IF vb <> va THEN IINeg(Q);
  IF d > 1 THEN IISetZero(BN); {Speicher wieder freigeben}
END;{IIDiv}
 
{.A+}
PROCEDURE IIShiftRight(VAR X:InfIntT;S:longint);
VAR q : InfIntT;
BEGIN
  REPEAT
    IF X <> nil THEN
    BEGIN
      q := X^.next;
      dispose(X);
      X := q;
    END
    ELSE S :=  0;
    S := pred(S);
  UNTIL S <= 0;
END;{IIShiftRight}
 
PROCEDURE IIShiftLeft(VAR X:InfIntT;S:longint);
VAR q : InfIntT;
BEGIN
  REPEAT
    new(q);
    q^.z := 0;
    q^.next := X;
    X  := q;
    S := pred(S);
  UNTIL S <= 0;
END;{IIShiftLeft}
 
FUNCTION IIVergleich(X,Y:InfIntT):integer;
VAR Xp,Yp : InfIntT;
    v ,h  : longint;
BEGIN
  v := 0;
  IF X = Y THEN IIVergleich := 0
  ELSE
  BEGIN
    Xp := X; Yp := Y;
    WHILE (Xp <> nil) AND (Yp <> nil) DO
    BEGIN
      h := Xp^.z - Yp^.z;
      IF h <> 0 THEN v := h;
      Xp := Xp^.next;
      Yp := Yp^.next;
    END;
    IF Xp <> nil THEN IIVergleich := IISign(X)
    ELSE IF Yp <> nil THEN IIVergleich := - IISign(Y)
         ELSE IF v = 0 THEN IIVergleich := 0
              ELSE IIVergleich := Sign(v);
  END;
END;{IIVergleich}

{*****************************************************************************}
{*                           Ein- und Ausgabe                                *}
{*****************************************************************************}

PROCEDURE DivStrInt(A:PChar;B:longint;VAR Q:PChar;VAR R:longint);
{ Dividiert einen String mit einer positiven ganzen Zahl mit Rest durch B   }
  VAR k  : longint;
      FC : integer;
      Ar    : comp;    {Damit auch durch longints > 3600 geteilt werden kann}
      t  : ARRAY[0..20]OF char;
BEGIN
  k := trunc(ln(B)/ln(10));  {Stellenzahl von B                             }
  val(StrLCopy(t,A,k),AR,FC);
  StrCopy(Q,'');
  k := succ(k);
  IF (AR < B) AND (StrLen(A) >= k) THEN
  BEGIN
    AR := AR * 10 + (ord(A[pred(k)])-ord('0'));
    k := succ(k);
  END;
  WHILE k <= StrLen(A) DO
  BEGIN
    StrPCopy(t,chr(trunc(Ar/B)+ord('0')));
    StrCat(Q,t);
    R := trunc(AR - B * 1.0 * trunc(Ar/B));
    AR := R * 10.0 + (ord(A[pred(k)])-ord('0'));
    k := succ(k);
  END;
  StrPCopy(t,chr(trunc(Ar/B)+ord('0')));
  StrCat(Q,t);
  R := trunc(AR - 1.0 * B * trunc(Ar/B));
END;{DivStrInt}


FUNCTION StringtoInfInt(S:PChar;VAR x:InfIntT):boolean;
{Wandelt einen String in eine Liste um. Ergibt false wenn ein Fehler auftritt}
  VAR k,v : longint;
      p   : InfIntT;
      q,r : PChar;
      len : integer;
BEGIN
  len := StrLen(s)+1;
  getmem(q,Len);
  getmem(r,Len);
  StrCopy(r,s);
  IF r[0] = '-' THEN
  BEGIN
    v := -1;
    StrMove(@r[1],r,StrLen(r)-1);
  END
  ELSE v := 1;
  FOR k := 0 TO pred(StrLen(r)) DO
    IF NOT (r[k] IN ['0'..'9']) THEN
    BEGIN
      StringtoInfInt:=false;
      freemem(q,StrLen(s)+1);
      freemem(r,StrLen(s)+1);
      exit
    END;
  IF (r[0] = '0') AND (StrLen(r) = 1) THEN
  BEGIN
    x := nil;
    freemem(q,Len);
    freemem(r,Len);
    exit
  END
  ELSE
  BEGIN
    DivStrInt(r,Basis,q,k);
    StrCopy(r,q);
    new(X);
    x^.z := v*k;
    x^.next := nil;
    p := x;
  END;
  WHILE (StrLen(r) > 1) OR (r[0] <> '0') DO
  BEGIN
    DivstrInt(r,Basis,q,k);
    StrCopy(r,q);
    new(p^.next);
    p := p^.next;
    p^.z := v*k;
    p^.next := nil;
  END;
  freemem(q,Len);
  freemem(r,Len);

END;{StringToInfInt}


PROCEDURE WriteInfInt1(x:InfIntT);{nur fr Basen 10^k}
  VAR p : InfIntT;
      r : longint;
      s : string[5];
BEGIN
  IIInvert(x);
  p := x;
  IF p = nil THEN write('0')
  ELSE BEGIN write(p^.z); p := p^.next END;
  WHILE p <> nil DO
  BEGIN
    r := abs(p^.z);
    str(r+Basis,s);
    delete(s,1,1);
    write(s);
    p := p^.next;
  END;
  IIInvert(x)
END; {WriteInfInt1}

{.A-}
PROCEDURE WriteInfInt2(VAR Dev:text;X:InfIntT);
TYPE LiFo  = ^Digit;
     Digit = RECORD
               D : longint;
               Link : Lifo;
             END;
VAR DigitStack : Lifo;
    v,d        : longint;
    s          : string[4];
    R,R2       : InfIntT;
  PROCEDURE StoreDigit(D:longint);
  VAR p : LiFo;
  BEGIN
    new(p);
    p^.Link := DigitStack;
    DigitStack := p;
    p^.D := D;
  END;{StoreDigit}

  FUNCTION GetDigit:longint;
  VAR p : LiFo;
  BEGIN
    IF DigitStack <> nil THEN
    BEGIN
      GetDigit := DigitStack^.D;
      p := DigitStack;
      DigitStack := DigitStack^.Link;
      dispose(p);
    END;
  END;{GetDigit}

BEGIN {WriteInfInt2}
  DigitStack := nil;
  v := IISign(X);
  R := IICopy(X);
  IF v < 0 THEN IINeg(R);
  WHILE R <> nil DO
  BEGIN
    IISDiv(R,1000,R2,d);
    Storedigit(d);
    IISetZero(R);
    R := R2;
  END;
  IF DigitStack = nil THEN write(Dev,'0')
  ELSE
  BEGIN
    write(Dev,v * GetDigit);
    WHILE DigitStack <> nil DO
    BEGIN
      str(GetDigit+1000,s);
      delete(s,1,1);
      write(Dev,s:4);
    END;
  END;
END; {WriteInfInt2}

{.A+}
PROCEDURE WriteInfInt(X:InfIntT);
BEGIN
  WriteInfInt2(Output,X);
END;{WriteInfInt}

{*****************************************************************************}
{*                          Zahlenumwandlungen                               *}
{*****************************************************************************}

FUNCTION InttoInfInt(x:longint):InfIntT;
{Verwandelt eine normale longint in eine InfIntT                             }
  VAR q : InfIntT;
      v : longint;
BEGIN
  IF x = 0 THEN InttoInfInt := nil
  ELSE
  BEGIN
    IF x < 0 THEN v := -1 ELSE v := 1;
    x := abs(x);
    new(q);
    q^.z := v * (x mod Basis);
    q^.next := nil;
    InttoInfInt := q;
    x := x DIV Basis;
    WHILE x > 0 DO
    BEGIN
      new(q^.next);
      q := q^.next;
      q^.next := nil;
      q^.z := v * (x MOD Basis);
      x := x DIV Basis;
    END;
  END;
END;{IntToInfInt}

FUNCTION compToInfInt(R:comp):InfIntT;
VAR h : comp;
    Erg,Ep : InfIntT;
    v : longint;
BEGIN
  IF abs(R) < 1 THEN Erg := nil
  ELSE
  BEGIN
    new(Erg); Ep := Erg;
    IF R < 0 THEN v := -1 ELSE v := 1;
    R := abs(R);
    h := int(R/Basis);
    Ep^.z := v * round(R - h*Basis);
    R := h;
    WHILE R <> 0 DO
    BEGIN
      new(Ep^.next); Ep := Ep^.next;
      h := int(R/Basis);
      Ep^.z := v * round(R - h*Basis);
      R := h;
    END;
    Ep^.next := nil;
  END;
  compToInfInt := Erg;
END;{compToInfInt}

FUNCTION InfIntTocomp(X:InfIntT):comp; {Achtung: keine Prfung auf Bereichsber-}
VAR r : comp;                         {schreitung                              }
    Xp : InfIntT;
BEGIN
  IIInvert(X);
  Xp := X;
  r  := 0;
  WHILE Xp <> nil DO
  BEGIN
    r := r * Basis + Xp^.z;
    Xp := Xp^.next;
  END;
  IIInvert(X);
  InfIntTocomp :=  r;
END;{InfIntTocomp}
 
FUNCTION InfIntToInt(X:InfIntT):longint;
VAR r : comp;
BEGIN
  InfIntToInt := 0;
  r := InfIntTocomp(X);
  IF abs(r) <= 2147483648.0 THEN InfIntToInt := trunc(R)
  ELSE writeln(#7,'Fehler: InfIntT konnte nicht in Integer verwandelt werden!');
END;{InfIntToInt}
 
{*****************************************************************************}
{*                    Weitere Operationen mit InfInts                        *}
{*****************************************************************************}
 
PROCEDURE IIInc(VAR X:InfIntT;S:longint);
VAR Sp,Erg : InfIntT;
BEGIN
  Sp := IntToInfInt(S);
  Erg := IISum(X,Sp);
  IISetzero(Sp);
  IISetZero(X);
  X := Erg;
END; {IIInc}
 
 
FUNCTION IIPot(X:InfIntT;Y: longint):InfIntT;
{Berechnet X^Y durch sukkzessives Multiplizieren und Quadrieren}
VAR H1,H2 : InfIntT;
BEGIN
  IF Y = 1 THEN  IIPot := IICopy(X)
  ELSE
    IF odd(Y) THEN
    BEGIN
      H1 := IIPot(X,Y-1);
      IIPot := IIProd(H1,X);
      IISetZero(H1);
    END
    ELSE
      IF Y = 2 THEN  IIPot := IIProd(X,X)
      ELSE
      BEGIN
        H1 := IIPot(X,Y DIV 2);
        IIPot := IIProd(H1,H1);
        IISetZero(H1);
      END;
END; {IIPot}
 
FUNCTION IIGGT(A,B : InfIntT):InfIntT;
{Berechnet den ggT von A und B mit dem Euklid-Algorithmus}
VAR A1,B1,Q,R : InfIntT;
BEGIN
  A1 := IIcopy(A); B1 := IICopy(B);
  IF IISign(A1) < 0 THEN IINeg(A1);
  IF IISign(B1) < 0 THEN IINeg(B1);
  REPEAT
    IIDiv(A1,B1,Q,R);
    IISetZero(Q); {wird nicht gebraucht daher Speicher wieder freigeben}
    IISetZero(A1);{wird auch nicht mehr gebraucht}
    A1 := B1; B1 := R;
  UNTIL B1 = nil;
  IIGGT := A1;
END; {GGT}
 
FUNCTION IIWurzel(X:InfIntT):InfIntT; {Gibt den ganzzahligen Teil der Quadrat-
  wurzel von X zurck }
VAR D2,R,Q,H,Xp,W    : InfIntT;
    k,ns             : longint;
    mr               : comp;
    korrekt          : boolean;

  FUNCTION Schaetzung(A,B:InfIntT):longint;
  {ersetzt die Division im Wurzelalgorithmus. Liefert einen Schtzwert fr die
   erste Stelle des Quotienten. Dieser Schtzwert ist nie zu gross. }
  VAR R,Rp,Bp,BN,RP2,Qp :InfIntT;
      mr : comp;
      d,b1,b2,qd : longint;
  BEGIN
    IF B = nil THEN
    BEGIN
      writeln(#7,'Fehler: Division durch 0');
      exit;
    END;
    IF A = nil THEN
    BEGIN
      Schaetzung := 0;
      exit;
    END;
    IF B^.next = nil THEN
    BEGIN {Lnge von B = 1 -> einfacher Algorithmus}
      {A ist dann auch hchstens 2-stellig}
      Schaetzung := trunc(InfIntTocomp(A)/InfIntTocomp(B));
      exit;
    END;
    R := IICopy(A);
    {hier beginnt der eigentliche Divisions-Algorithmus}
    {1. Normalisierung}
    IIInvert(B);
    d := Basis DIV (B^.z +1);
    IIInvert(B);
    IF d <> 1 THEN
    BEGIN
      BN := IISProd(B,d);
      Rp := IISProd(R,d);
      IISetZero(R); R := Rp;
    END
    ELSE BN := B;
    IIInvert(R);IIInvert(BN);
    IF BN^.z <= R^.z THEN
    BEGIN
      RP := R;
      new(R);
      R^.z := 0;
      R^.next := RP;
    END;
    {Ende der Normalisierung,Dividend R und Divisor BN sind invertiert }
    RP2 := R;
    BP := BN;
    REPEAT RP2 := RP2^.next; Bp := Bp^.next UNTIL (Bp = nil) OR (RP2 = nil);
    IF Rp2 = nil THEN
    BEGIN {In diesem Fall ist A < B}
      IF d > 1 THEN IISetZero(BN) ELSE IIInvert(BN);
      IISetZero(R);
      Schaetzung := 0;
      exit;
    END;
    b1 := BN^.z;b2 := Bn^.next^.z; IIInvert(BN);
    {BN wieder in der normalen Reihenfolge }
    {Quotientenziffer schtzen}
    mr := 1.0 * Basis * R^.z + R^.next^.z; {die beiden hchsten Ziffern}
    IF R^.z >= B1+1 THEN qd := Basis - 1
    ELSE qd := trunc(mr/(B1+1));
    IISetZero(R);
    Schaetzung := qd;
    IF d > 1 THEN IISetZero(BN);
  END;{Schaetzung}
 
BEGIN
  IF X = nil THEN IIWurzel := nil
  ELSE
  IF IISign(X) < 0 THEN
  BEGIN
    writeln(#7,'Fehler: Negativer Radikand');
    IIWurzel := nil;
  END
  ELSE
  BEGIN {hier ist X sicher > 0}
    R := nil; D2 := nil;
    {1. Feststellen ob X gerade oder ungerade Stellenzahl hat}
    Xp := X;
    k := 0;
    WHILE Xp <> nil DO
    BEGIN
      k := succ(k) AND 1;
      Xp := Xp^.next;
    END;
    { Wenn k = 0 dann hat X gerade Stellenzahl }
    IIInvert(X); {Reihenfolge der Ziffern nach absteigendem Grad}
    {1. Stelle des Resultates bestimmen }
    IF k = 0 THEN
    BEGIN {1. und 2. Stelle von X verwenden}
      mr := 1.0 * X^.z * Basis + X^.next^.z;
      Xp := X^.next^.next;
    END
    ELSE {nur 1. Stelle verwenden }
    BEGIN
      mr := X^.z;
      Xp := X^.next;
    END;
    ns := trunc(Sqrt(mr));
    new(W); W^.next := nil; W^.z := ns;
    R := compToInfInt(mr-1.0*ns*ns); {1.Teilrest}
    WHILE Xp <> nil DO {weitere Stellen bestimmen}
    BEGIN
      D2 := IISProd(W,2*Basis);
      IIInc(D2,Basis); {D2 := Basis*2*W +Basis}
      {nchste zwei Stellen zu R herunternehmen}
      IIShiftleft(R,2);
      R^.next^.z := Xp^.z;
      Xp := Xp^.next;
      R^.z := Xp^.z;
      Xp := Xp^.next;
      {Schtzung fr die nchste Stelle berechnen}
      ns := Schaetzung(R,D2);
      {neuen Teilrest berechnen}
      IIInc(D2,-Basis); {D2:= 2*Basis*W)}
      H := IISProd(D2,ns);
      Q := IIDiff(R,H); IISetzero(R);IISetZero(H);
      H := compToInfInt(1.0*ns*ns);
      R := IIDiff(Q,H); IISetZero(Q); IISetzero(H);
      korrekt := false;
      {Schtzung korrigieren bis sie stimmt}
      REPEAT
        IIInc(D2,2*ns); {D2 := 2*(Basis*W+ns)}
        IF IIVergleich(R,D2) <= 0 THEN korrekt := true
        ELSE
        BEGIN
          IIInc(D2,Basis);
          IF IIVergleich(R,D2) <= 0 THEN k := 1
          ELSE
          BEGIN
            IIDiv(R,D2,Q,H); IISetZero(H);
            k := InfIntToInt(Q); IISetZero(Q);
          END;
          IIInc(D2,-Basis);IIInc(D2,-2*ns);
          H := IISProd(D2,k);
          Q := IIDiff(R,H);
          IISetZero(R); IISetZero(H);
          H := compToInfInt(2.0*k*ns+1.0*k*k);
          R := IIDiff(Q,H); IISetZero(Q); IISetZero(H);
          ns := ns + k;
        END;
      UNTIL korrekt;
      IIShiftLeft(W,1);
      W^.z := ns;
    END; {WHILE }
    IIInvert(X); IISetZero(R); IISetZero(D2);
    IIWurzel := W;
  END;{ELSE}
END; {IIWurzel}

{***************************END OF INFINT.BIB*********************************}

  CONSTRUCTOR TInfInt.Init;
  BEGIN
    number := nil;
  END;

  CONSTRUCTOR TInfInt.AssignStr(s:PChar);
  BEGIN
    IF NOT StringToInfInt(s,number) THEN
      fail;
  END;

  CONSTRUCTOR TInfInt.Assign(X:PInfInt);
  BEGIN
    number := IICopy(X^.number);
  END;

  CONSTRUCTOR TInfInt.AssignLong(X:Longint);
  BEGIN
    number := IntToInfInt(X);
  END;

  CONSTRUCTOR TInfInt.AssignComp(X:comp);
  BEGIN
    number := compToInfInt(X);
  END;

  CONSTRUCTOR TInfInt.Sum(X,Y:PInfInt);
  BEGIN
    number := IISum(X^.number,Y^.number);
  END;

  CONSTRUCTOR TInfInt.Diff(X,Y:PInfInt);
  BEGIN
    number := IIDiff(X^.number,Y^.number);
  END;

  CONSTRUCTOR TInfInt.Prod(X,Y:PInfint);
  BEGIN
    number := IIProd(X^.number,Y^.number);
  END;

  CONSTRUCTOR TInfInt.Quot(X,Y:PInfInt);
    VAR
      r : InfIntT;
  BEGIN
    IIDiv(X^.number,Y^.Number,number,r);
    IISetZero(r);
  END;

  CONSTRUCTOR TInfInt.Remainder(X,Y:PInfInt);
    VAR
      q : InfIntT;
  BEGIN
    IIDiv(X^.number,Y^.Number,q,number);
    IISetZero(q);
  END;

  CONSTRUCTOR TInfInt.GGT(X,Y:PInfInt);
  BEGIN
    number := IIGGT(X^.number,Y^.number);
  END;

  DESTRUCTOR  TInfInt.Done;
  BEGIN
    IF number <> nil THEN
      IISetZero(number);
    inherited Done;
  END;

  PROCEDURE TInfInt.SetZero;
  BEGIN
    IF number <> nil THEN
      IISetZero(number);
  END;

  FUNCTION TInfInt.Sign:integer;
  BEGIN
    Sign := IISign(number);
  END;

  PROCEDURE TInfInt.ChangeSign;
  BEGIN
    IINeg(number);
  END;

  PROCEDURE TInfInt.AssignNew(X:PinfInt);
  BEGIN
    IISetZero(number);
    number := IICopy(X^.number);
  END;

  PROCEDURE TInfInt.Add(X:PInfInt);
    VAR t : InfIntT;
  BEGIN
    t := IISum(number,X^.number);
    IISetZero(number);
    number := t;
  END;

  PROCEDURE TInfInt.Sub(X:PInfInt);
    VAR t : InfIntT;
  BEGIN
    t := IIDiff(number,X^.number);
    IISetZero(number);
    number := t;
  END;

  PROCEDURE TInfInt.Multiply(X:PInfInt);
    VAR t : InfIntT;
  BEGIN
    t := IIProd(number,X^.number);
    IISetZero(number);
    number := t;
  END;

  FUNCTION TInfInt.DivideRemainder(Denom:PinfInt): PInfInt;
    VAR
      S : InfIntT;
      T : PInfInt;
  BEGIN
    new(T,Init);
    IIDiv(number,Denom^.number,S,T^.number);
    IISetZero(number);
    number := s;
    DivideRemainder := T;
  END;

  PROCEDURE TInfInt.MultiplyLong(X:longInt);
    VAR t : InfIntT;
  BEGIN
    t := IISProd(number,X);
    IISetZero(number);
    number := t;
  END;


  FUNCTION TInfInt.DivideRemainderLong(X:longint):longint;
    VAR
      S : InfIntT;
      T : Longint;
  BEGIN
    IISDiv(number,X,S,T);
    IISetZero(number);
    number := s;
    DivideRemainderLong := T;
  END;

  PROCEDURE TInfInt.ShiftLeft(n:longint);
  BEGIN
    IIShiftLeft(number,n);
  END;

  PROCEDURE TInfInt.ShiftRight(n:longint);
  BEGIN
    IIShiftRight(number,n);
  END;

  PROCEDURE TInfInt.Power(exp:longint);
    VAR
      T : InfIntT;
  BEGIN
    T := IIPot(number,exp);
    IISetZero(number);
    number := T;
  END;

  PROCEDURE TInfInt.Sqrt;
    VAR
      T : InfIntT;
  BEGIN
    T := IIWurzel(number);
    IISetZero(number);
    number := T;
  END;

  FUNCTION TInfInt.Compare(X:PInfInt):integer;
  BEGIN
    Compare := IIVergleich(number,X^.number);
  END;

  FUNCTION TInfInt.Str:PChar;
    VAR
      i,j : longint;
      p : InfIntT;
      s : PChar;
      u : ARRAY[0..9] OF Char;
  BEGIN
    i := 0;
    p := number;
    WHILE p <> nil DO
    BEGIN
      p := p^.next;
      inc(i);
    END;
    IF i = 0 THEN
    BEGIN
      GetMem(s,2);
      s := '0';
    END
    ELSE
    BEGIN
      i := (i * 8) + 1;
      IF i > 64000 THEN Halt(99);
      Getmem(s,i);
      IF IISign(number) = -1 THEN s := '-' ELSE s := ' ';
      IIInvert(number);
      p := number;
      i := 1;
      WHILE p <> nil DO
      BEGIN
        system.str(abs(p^.z),u);
        j := 0;
        WHILE u[j] <> #0 DO
        BEGIN
          s[i] := u[j];
          inc(i);
          inc(j);
        END;
        p := p^.next;
      END;
      s[j] := #0;
    END;
    Str := s;
  END;

  PROCEDURE TInfInt.Write(VAR F:Text);
    VAR
      p : InfIntT;
  BEGIN
    IF number <> nil THEN
    BEGIN
      IF Sign < 0 THEN
        system.write(F,'-');
      IIInvert(number);
      p := number;
      WHILE p <> nil DO
      BEGIN
        system.write(f,abs(p^.z));
        p := p^.next;
      END;
      IIInvert(number);
    END
    ELSE
      system.write(f,'0');
  END;

  CONSTRUCTOR TInfInt.Load(VAR S:TStream);
    VAR Xp : InfIntT;
        i  : longint;
  BEGIN
    S.read(i,SizeOf(i));
    IF i = $FFFFFFFF THEN number := nil
    ELSE
    BEGIN
      new(number); Xp := number;
      Xp^.z := i;
      S.read(i,SizeOf(i));
      WHILE i <> $FFFFFFFF DO
      BEGIN
        new(Xp^.next);
        Xp := Xp^.next;
        Xp^.z := i;
        S.read(i,SizeOf(i));
      END;
      Xp^.next := nil;
    END;
  END;

  PROCEDURE TInfInt.Store(VAR S:TStream);
    VAR Xp : InfIntT; i : longint;
  BEGIN
    Xp := number;
    WHILE Xp <> nil DO
    BEGIN
      S.write(Xp^.z,SizeOf(Xp^.z));
      Xp := Xp^.next;
    END;
    i := $FFFFFFFF;
    S.Write(i,SizeOf(i)); {Als Schlussmarkierung}
  END;

  PROCEDURE RegisterInfInt;
  BEGIN
    RegisterType(RInfInt);
  END;

END.