UNIT machine;

INTERFACE

USES opcodes,table,manager,analyze,putexpr,whatever;

	procedure eval;

IMPLEMENTATION

	procedure eval;

	var
		nextexp: expPTR;

		{ ----- constant registers ----- }

		_omega,_true: expPTR;

		{ ----- scratch registers -----}

		x,y,z,a,b,ab,loc,c1,c2,e1,v,cx,recipe,tf: expPTR;

		{ ----- garbage collection ----- }

		procedure collect_garbage;
		var
			marked,collected: integer;

			procedure markexp(e:expPTR);
			begin {markexp}
				if NOT e^.in_use then
					begin
						e^.in_use:= true;
						inc(marked);
						if ispair(e) then
							begin
								markexp(car(e));
								markexp(cdr(e))
							end
					end;
			end {markexp};

			procedure collect;
			var
				current,previous: expPTR;
			begin {collect}
				if root = nil then
					begin
						writeln('MEMORY CORRUPTED!');
						HALT
						end
				else
					begin {collecting}
						previous:= root;
						current:= root^.uplink;
						while current <> nil do
								if NOT current^.in_use then
									begin
										previous^.uplink:= current;
										current^.downlink:= previous;
										previous:= current;
										current:= current^.uplink;
										inc(collected)
									end
								else
									current:= current^.uplink;
						previous^.uplink:= current; { guaranteed = nil }
						ff:= previous; { guaranteed <> nil }
					end {collecting}
			end{ collect};

		begin {collect_garbage}
			marked:=0;
			{ mark constant registers }
			markexp(_nada);
			markexp(_omega);
			markexp(_true);
			{ mark main registers }
			markexp(s);
			markexp(e);
			markexp(c);
			markexp(d);
			markexp(r);
			{ mark scratch registers }
			markexp(x);
			markexp(y);
			markexp(z);
			markexp(a);
			markexp(b);
			markexp(ab);
			markexp(loc);
			markexp(c1);
			markexp(c2);
			markexp(e1);
			markexp(v);
			markexp(cx);
			markexp(recipe);
			markexp(tf);
			writeln('marked    = ',marked);
			collected:= 0;
			collect;
			writeln('collected = ',collected);
			writeln('total     = ',marked+collected);
			putsepr('');
		end {collect_garbage};

		{ ----- pair constructor ----- }

		function conspair(a,b:expPTR):expPTR;
		begin {conspair}
			if ff = root then collect_garbage;
			if ff = root then
				begin
					writeln('[conspair] OUT OF MEMORY!');
					HALT
				end
			else
				begin
					conspair:= ff;
					ff:= ff^.downlink;
					conspair^.exptype:= _pair;
					conspair^.car:= a;
					conspair^.cdr:= b
				end
		end {conspair} ;

		{ ----- number constructor ----- }

		function consnumber(z:longint):expPTR;
		begin {consnumber}
			if ff = root then collect_garbage;
			if ff = root then
				begin
					writeln('[consnumber] OUT OF MEMORY!');
					HALT
				end
			else
				begin
					consnumber:= ff;
					ff:= ff^.downlink;
					consnumber^.exptype:= _number;
					consnumber^.value:= z
				end
		end {consnumber} ;

		{ ----- pair reconstructors ----- }

		procedure setcar(var e:expPTR; v:expPTR);
		begin {setcar}
			if ispair(e) then
				e^.car:= v
			else
				begin
					writeln('[setcar] target not _pair!');
					HALT
			  end
		end {setcar} ;

	  procedure setcdr(var e:expPTR; v:expPTR);
		begin {setcdr}
			if ispair(e) then
			  e^.cdr:= v
			else
				begin
					writeln('[setcdr] target not _pair!');
					HALT
				end
		end {setcdr} ;

		{ ----- functions ----- }

		function eq(x,y:expPTR):boolean;
		begin {eq}
			if issymbol(x) and issymbol(y) then
				eq:= ( index(x) = index(y) )
			else if isnumber(x) and isnumber(y) then
				eq:= ( value(x) = value(y) )
			else if issign(x) and issign(y) then
				eq:= ( key(x) = key(y) )
			else if ispair(x) and ispair(y) then
				eq:= eq( car(x) , car(y) )
							 and
						 eq( cdr(x) , cdr(y) )
			else
				eq:= false
		end {eq} ;

		function lt(x,y:expPTR):boolean;
		begin {lt}
			if isnumber(x) and issymbol(y) then
				lt:= true
			else if isnumber(x) and isnumber(y) then
				lt:= ( value(x) < value(y) )
			else if issymbol(x) and issymbol(y) then
				lt:= ( nametable[index(x)] < nametable[index(y)] )
			else
				lt:= false
		end {lt} ;

		function do_add(x,y:expPTR):expPTR;
		var temp: longint;
		begin {do_add}
			if isnumber(x) and isnumber(y) then
				begin
					temp:= value(x) + value(y);
					do_add:= consnumber(temp)
				end
			else
				begin
					writeln('[do_add] arg(s) not _number(s)!');
					HALT
				end
		end {do_add} ;

		function do_sub(x,y:expPTR):expPTR;
		var temp: longint;
		begin {do_sub}
			if isnumber(x) and isnumber(y) then
				begin
					temp:= value(x) - value(y);
					do_sub:= consnumber(temp)
				end
			else
				begin
					writeln('[do_sub] arg(s) not _number(s)!');
					HALT
				end
		end {do_sub} ;

		function do_mul(x,y:expPTR):expPTR;
		var temp: longint;
		begin {do_mul}
			if isnumber(x) and isnumber(y) then
				begin
					temp:= value(x) * value(y);
					do_mul:= consnumber(temp)
				end
			else
				begin
					writeln('[do_mul] arg(s) not _number(s)!');
					HALT
				end
		end {do_mul} ;

		function do_div(x,y:expPTR):expPTR;
		var temp: longint;
		begin {do_div}
			if isnumber(x) and isnumber(y) then
				if value(y) <> 0 then
					begin
						temp:= value(x) div value(y);
						do_div:= consnumber(temp)
					end
				else
					begin
						writeln('[do_div] ZERO DIVIDE ATTEMPT!');
						HALT
					end
			else
				begin
					writeln('[do_div] arg(s) not _number(s)!');
					HALT
				end
		end {do_div} ;

		{ ----- interpret variables ----- }

		function fetch(ij:expPTR; e:expPTR):expPTR;

			function get(n:integer; s:expPTR):expPTR;
			begin {get}
				if n = 0 then
					get:= car(s)
				else
					get:= get( n-1 , cdr(s) )
			end {get} ;

		begin {fetch}
			fetch:= get( value(cdr(ij)) , get( value(car(ij)) , e ) )
		end {fetch} ;

	begin {eval}

		{ ----- initialize constant registers ----- }

		_omega:= makesymbol('_omega');
		_true:= makesymbol('_non_nil');

		{ ----- initialize scratch registers ----- }

		x:= _nada;
		y:= _nada;
		z:= _nada;
		a:= _nada;
		b:= _nada;
		ab:= _nada;
		loc:= _nada;
		c1:= _nada;
		c2:= _nada;
		e1:= _nada;
		v:= _nada;
		cx:= _nada;
		recipe:= _nada;
		tf:= _nada;

		{ main loop }

		repeat

			nextexp:= car(c);

			c:= cdr(c); { distributes... }

			if exptype(nextexp) = _sign then
				begin
					cmd:= key(nextexp);
					histo[cmd]:= histo[cmd] + 1
				end
			else
				begin
					writeln('[eval] ord(exptype(nextexp)) = ',
												ord(exptype(nextexp)),' <> ord(_sign)!');
					HALT
				end;

			case cmd of

			_ld: begin
				loc:= car(c);
				x:= fetch(loc,e);
				s:= conspair(x,s);
				c:= cdr(c);
				{ snip }
        loc:= _nada;
        x:= _nada
			end;

			_ldc: begin
				x:= car(c);
				s:= conspair(x,s);
				c:= cdr(c);
				{ snip }
				x:= _nada
			end;

			_sel: begin
				x:= car(s);
				if isnada(x) then
					cx:= car(cdr(c))
				else
					cx:= car(c);
				s:= cdr(s);
				d:= conspair(cdr(cdr(c)),d);
				c:= cx;
				{ snip }
				x:= _nada;
				cx:= _nada
			end;

			_cons: begin
				a:= car(s);
				b:= car(cdr(s));
				s:= conspair(conspair(a,b),cdr(cdr(s)));
				{ snip }
				a:= _nada;
				b:= _nada
			end;

			_car: begin
				ab:= car(s);
				s:= conspair(car(ab),cdr(s));
				{ snip }
				ab:= _nada
			end;

			_cdr: begin
				ab:= car(s);
				s:= conspair(cdr(ab),cdr(s));
				{ snip }
				ab:= _nada
			end;

			_isnada: begin
				if isnada(car(s)) then
					s:= conspair(_true,cdr(s))
				else
					s:= conspair(_nada,cdr(s))
				end;

			_ispair: begin
				if ispair(car(s)) then
					s:= conspair(_true,cdr(s))
				else
					s:= conspair(_nada,cdr(s))
			end;

			_issymbol: begin
				if issymbol(car(s)) then
					s:= conspair(_true,cdr(s))
				else
					s:= conspair(_nada,cdr(s))
				end;

			_isnumber: begin
				if isnumber(car(s)) then
					s:= conspair(_true,cdr(s))
				else
					s:= conspair(_nada,cdr(s))
				end;

			_issign: begin
				if issign(car(s)) then
					s:= conspair(_true,cdr(s))
				else
					s:= conspair(_nada,cdr(s))
				end;

		 _isatom: begin
				if ispair(car(s)) then
					s:= conspair(_nada,cdr(s))
				else
					s:= conspair(_true,cdr(s))
		 end;

			_eq: begin
				y:= car(s);
				x:= car(cdr(s));
				if eq(x,y) then
					s:= conspair(_true,cdr(cdr(s)))
				else
					s:= conspair(_nada,cdr(cdr(s)));
				{ snip }
				x:= _nada;
				y:= _nada
			end;

			_lt: begin
				y:= car(s);
				x:= car(cdr(s));
				if lt(x,y) then
					s:= conspair(_true,cdr(cdr(s)))
				else
					s:= conspair(_nada,cdr(cdr(s)));
				{ snip }
				x:= _nada;
				y:= _nada
			end;

			_join: begin
				c:= car(d);
				d:= cdr(d)
			end;

			_rtn: begin
				x:= car(s);
				s:= conspair(x,car(d));
				e:= car(cdr(d));
				c:= car(cdr(cdr(d)));
				d:= cdr(cdr(cdr(d)));
				{ snip }
				x:= _nada
			end;

			_ldf: begin
				c1:= car(c);
				s:= conspair(conspair(c1,e),s);
				c:= cdr(c);
				{ snip }
				c1:= _nada
			end;

			_ap:	begin
				c1:= car(car(s));
				e1:= cdr(car(s));
				v:= car(cdr(s));
				s:= cdr(cdr(s));
				d:= conspair(s,conspair(e,conspair(c,d)));
				s:= _nada;
				e:= conspair(v,e1);
				c:= c1;
				{ snip }
        c1:= _nada;
				e1:= _nada;
				v:= _nada
			end;

			_dum: begin
				e:= conspair(_omega,e);
			end;

			_rap: begin
				c1:= car(car(s));
				e1:= cdr(car(s));
				v:= car(cdr(s));
				setcar(e1,v);
				d:= conspair(cdr(cdr(s)),conspair(cdr(e),conspair(c,d)));
				s:= _nada;
				c:= c1;
				{ snip }
				c1:= _nada;
        e1:= _nada;
				v:= _nada
			end;

			_sor: begin
				c1:= car(c);
				c2:= car(cdr(c));
				c:= cdr(cdr(c));
				d:= conspair(c,d);
				r:= conspair(s,conspair(e,conspair(c2,conspair(d,r))));
				c:= c1;
				{ snip }
        c1:= _nada;
				c2:= _nada
			end;

			_non: begin
				if isnada(r) then
					begin
						writeln('[eval] terminal _NON implies _STOP.');
						cmd:= _stop
					end
				else
					begin
						s:= car(r);
						e:= car(cdr(r));
						c:= car(cdr(cdr(r)));
						d:= car(cdr(cdr(cdr(r))));
						r:= cdr(cdr(cdr(cdr(r))));
					end
				end;

			_lde: begin
				recipe:= conspair(_nada,conspair(car(c),e));
				s:= conspair(recipe,s);
				c:= cdr(c)
			end;

			_ap0: begin
				recipe:= car(s);
				tf:= car(recipe);
				if isnada(tf) then
					begin
						d:= conspair(s,conspair(e,conspair(c,d)));
						s:= _nada;
						e:= cdr(cdr(recipe));
						c:= car(cdr(recipe))
					end
				else
					s:= conspair(cdr(recipe),cdr(s))
			end;

			_upd: begin
				x:= car(s);
				ab:= car(d);
				setcar(ab,conspair(_true,x));
				s:= conspair(x,cdr(ab));
				e:= car(cdr(d));
				c:= car(cdr(cdr(d)));
				d:= cdr(cdr(cdr(d)))
			end;

			_caar: s:= conspair(car(car(car(s))),cdr(s));

			_cadr: s:= conspair(car(cdr(car(s))),cdr(s));

			_cdar: s:= conspair(cdr(car(car(s))),cdr(s));

			_cddr: s:= conspair(cdr(cdr(car(s))),cdr(s));

			_cadar: s:= conspair(car(cdr(car(car(s)))),cdr(s));

			_caddr: s:= conspair(car(cdr(cdr(car(s)))),cdr(s));

			_cdddr: s:= conspair(cdr(cdr(cdr(car(s)))),cdr(s));

			_cadddr: s:= conspair(car(cdr(cdr(cdr(car(s))))),cdr(s));

			_cddddr: s:= conspair(cdr(cdr(cdr(cdr(car(s))))),cdr(s));

			_add: begin
				y:= car(s);
				x:= car(cdr(s));
				z:= do_add(x,y);
				s:= conspair(z,cdr(cdr(s)));
				{ snip }
        x:= _nada;
        y:= _nada;
        z:= _nada
			end;

			_sub: begin
				y:= car(s);
				x:= car(cdr(s));
				z:= do_sub(x,y);
				s:= conspair(z,cdr(cdr(s)));
				{ snip }
        x:= _nada;
        y:= _nada;
        z:= _nada
			end;

			_mul: begin
				y:= car(s);
				x:= car(cdr(s));
				z:= do_mul(x,y);
				s:= conspair(z,cdr(cdr(s)));
				{ snip }
        x:= _nada;
        y:= _nada;
        z:= _nada
			end;

			_div: begin
				y:= car(s);
				x:= car(cdr(s));
				z:= do_div(x,y);
				s:= conspair(z,cdr(cdr(s)));
				{ snip }
        x:= _nada;
        y:= _nada;
				z:= _nada
			end;

			_rplaca: begin
				a:= car(s);
				b:= car(cdr(s));
				setcar(a,b);
				s:= conspair(a,cdr(s));
				{ snip }
        a:= _nada;
        b:= _nada
			end;

			_bkpt: begin
				writeln('[eval] _BKPT reached.');
				a:= car(s);
				b:= car(cdr(s));
				putexp(a);write(' ');putexp(b);writeln;
				HALT
			end;

			_stop: writeln('[eval] _STOP reached.');

			else
				begin
					writeln('[eval] ord(cmd) = ',ord(cmd),' not implemented!');
					HALT
				end

			end {case}

		until ( cmd = _stop ) ;

	end {eval} ;

END.
