UNIT manager;

INTERFACE

USES Crt,opcodes,table,whatever;

type
	exptypes = (
		_symbol,
		_number,
		_sign,
		_pair);

	expPTR = ^exp;

	exp = record
		in_use: boolean;
		uplink,downlink: expPTR;
		case exptype:exptypes of
			_symbol: (index: symtableindex);
			_number: (value: longint);
			_sign: (key: opcode);
			_pair: (car,cdr: expPTR);
		end;

var
	{ ----- memory pointers ----- }

	_nada,ff,root: expPTR;

	{ ----- main registers -----}

	s,e,c,d,r: expPTR;

	{ ----- constructors ----- }

	function makesymbol(currtoken:symtabletype):expPTR;
	function makenumber(currtoken:symtabletype):expPTR;
	function makesign(currtoken:symtabletype):expPTR;
	function makepair:expPTR;

	{ ----- recognizers ----- }

	function issymbol(e:expPTR):boolean;
	function isnumber(e:expPTR):boolean;
	function issign(e:expPTR):boolean;
	function ispair(e:expPTR):boolean;
	function isnada(e:expPTR):boolean;

	{ ----- extractors ----- }

	function exptype(e:expPTR):exptypes;
	function index(e:expPTR):symtableindex;
	function value(e:expPTR):longint;
	function key(e:expPTR):opcode;
	function car(e:expPTR):expPTR;
	function cdr(e:expPTR):expPTR;

IMPLEMENTATION

	{ ----- memory management ----- }

	procedure showmemstats;
	begin {showmemstats}
		writeln('(available memory = ',MemAvail,')');
		writeln('(biggest free block ',MaxAvail,')')
	end {showmemstats};

	function heapfunc(size: word):integer; far;
	begin {heapfunc}
		if size <> 0 then
			begin
				putsepr('HeapFunc Status');
				showmemstats;
				putsepr('');
				heapfunc:= 1 { new returns nil }
			end
		end {heapfunc};

	procedure initmem;
	var
		current,previous: expPTR;
		uplinks,downlinks,verify: integer;
	begin {initmem}
		putsepr('initialize heap');
		showmemstats;
		new(root);
		if root <> nil then
			begin
				root^.downlink:= nil;
				uplinks:= 0;
				previous:= root;
				new(current);
				while current <> nil do
					begin
						previous^.uplink:= current;
						inc(uplinks);
						current^.downlink:= previous;
						current^.in_use:= false;
						previous:= current;
						new(current)
					end;
				previous^.uplink:= current; { guaranteed = nil }
				inc(uplinks);
				ff:= previous; { guaranteed <> nil }
				writeln('uplinks = ',uplinks);
				showmemstats;

				writeln('test downlinks.....');
				downlinks:= 0;
				current:= ff;
				while current <> nil do
					begin
						current:= current^.downlink;
						inc(downlinks)
					end;
				writeln('downlinks = ',downlinks);
				if downlinks <> uplinks then
					begin
						writeln('[initmem] downlinks <> uplinks!');
						HALT
					end;
				writeln('test uplinks...');
				verify:= 0;
				current:= root;
				while current <> nil do
					begin
						current:= current^.uplink;
						inc(verify)
					end;
				writeln('count = ',verify);
				if verify <> downlinks then
					begin
						writeln('[initmem] verify <> downlinks!');
						HALT
					end
			end
		else
			begin
				writeln('CAN''T INITIALIZE MEMORY!');
				HALT
			end;
		writeln('initialization complete.')
	end {initmem};

	{ ----- constructors ----- }

	function makeexp:expPTR;
	begin {makeexp}
		if ff = root then
			begin
				writeln('[makeexp] OUT OF MEMORY!');
				HALT
			end
		else
			begin
				makeexp:= ff;
				ff:= ff^.downlink
			end
	end {makeexp};

	function makesymbol(currtoken:symtabletype):expPTR;
	begin {makesymbol}
		makesymbol:= makeexp;
		makesymbol^.exptype:= _symbol;
		makesymbol^.index:= lookup(currtoken);
	end {makesymbol};

	function makenumber(currtoken:symtabletype):expPTR;
		function tointeger(currtoken:symtabletype):integer;
		var
			neg:boolean;
			s:integer;
			i,start:integer;
		begin {tointeger}
			neg:= ( currtoken[1] = '-' );
			s:= 0;
			if neg then start:= 2 else start:= 1;
			for i:= start to length(currtoken) do
				s:= 10*s + ( ord(currtoken[i]) - ord('0') );
			if neg then tointeger:= -s else tointeger:= s
		end {tointeger};
	begin {makenumber}
		makenumber:= makeexp;
		makenumber^.exptype:= _number;
		makenumber^.value:= tointeger(currtoken);
	end {makenumber};

	function makesign(currtoken:symtabletype):expPTR;
	begin {makesign}
		makesign:= makeexp;
		makesign^.exptype:= _sign;
		makesign^.key:= getopcode(currtoken);
	end {makesign};

	function makepair:expPTR;
	begin {makepair}
		makepair:= makeexp;
		makepair^.exptype:= _pair;
	end {makepair};

	{ ----- recognizers ----- }

	function issymbol(e:expPTR):boolean;
	begin
		issymbol:= ( exptype(e) = _symbol)
	end;
	function isnumber(e:expPTR):boolean;
	begin
		isnumber:= ( exptype(e) = _number )
	end;
	function issign(e:expPTR):boolean;
	begin
		issign:= ( exptype(e) = _sign )
	end;
	function ispair(e:expPTR):boolean;
	begin
		ispair:= ( exptype(e) = _pair )
	end;

	function isnada(e:expPTR):boolean;
	begin
		if not issymbol(e) then
			isnada:= false
		else
			isnada:= ( index(e) = index(_nada) )
	end;

	{ ----- extractors ----- }

	function exptype(e:expPTR):exptypes;
	begin {exptype}
		exptype:= e^.exptype
	end {exptype};

	function index(e:expPTR):symtableindex;
	begin {index}
		if issymbol(e) then
			index:= e^.index
		else
			begin
				writeln('[index] arg not _symbol!');
				HALT
			end
	end {index};

	function value(e:expPTR):longint;
	begin {value}
		if isnumber(e) then
			value:= e^.value
		else
			begin
				writeln('[value] arg not _number!');
				HALT
			end
	end {value};

	function key(e:expPTR):opcode;
	begin {key}
		if issign(e) then
			key:= e^.key
		else
			begin
				writeln('[key] arg not _sign!');
				HALT
			end
	end {key};

	function car(e:expPTR):expPTR;
	begin {car}
		if ispair(e) then
			car:= e^.car
		else
			begin
				writeln('[car] arg not _pair! ',ord(exptype(e)));
				HALT
			end
	end {car};

	function cdr(e:expPTR):expPTR;
	begin {cdr}
		if ispair(e) then
			cdr:= e^.cdr
		else
			begin
				writeln('[cdr] arg not _pair! ',ord(exptype(e)));
				HALT
			end
	end {cdr};

BEGIN

	 HeapError:= @heapfunc;

	 ClrScr;
	 initmem;

	 _nada:= makesymbol('nil')

END.