{*****************************************************************************}
{*       Kettenbruchentwicklungen von Quadratwurzeln von ganzen Zahlen       *}
{*       und Lsung der Pell'schen Gleichungen : X*X - D*Y*Y = 1            *}
{*                                                                           *}
{*       Verwendet die Unterprogrammsammlung InfIntT.BIB                      *}
{*---------------------------------------------------------------------------*}
{*       Autor :              C. Sperr                                       *}
{*       Datum :              02.07.1987                                     *}
{*       Letzte Revision :    20.07.1993                                     *}
{*****************************************************************************}

PROGRAM PELL;

  USES InfInt;

TYPE Kettenbruch = ^KBZiffer;
     KBZiffer    = RECORD
                     Ziffer : PInfInt;
                     next   : Kettenbruch;
                   END;
     Koerperelement = RECORD                {                           X*n^ +Y}
                         X,Y,Z : PInfInt;   { Bezeichnet Zahl der Form  --------}
                         N     : PInfInt;   {                               Z   }
                       END;

TYPE
  TCharArr = ARRAY[0..255] OF char;
  PCharArr = ^TCharArr;

VAR KB,Kptr    : Kettenbruch;
    N,X,Y,XX,YY: PInfInt;
    k,pl       : integer;
    entwickelt : boolean;
    S          : PChar;
    ch         : char;

PROCEDURE kuerzen(VAR K:Koerperelement);
{teilt alle Komponenten von K durch ihren GGT}
VAR GGT1,GGT2 : PInfInt;
BEGIN
  WITH K DO
  BEGIN
    GGT1 := new(PInfInt,GGT(X,Y));
    GGT2  := new(PInfInt,GGT(GGT1,Z));
    dispose(GGT1,done);
    dispose(X^.DivideRemainder(GGT2),Done);
    dispose(Y^.DivideRemainder(GGT2),Done);
    dispose(Z^.DivideRemainder(GGT2),Done);
    dispose(GGT2,Done);
  END;
END; {kuerzen}

PROCEDURE Kehrwert(VAR K:Koerperelement);
{berechnet den Kehrwert von K }
VAR X1,Y1,Z1,H1 : PInfInt;
BEGIN
  WITH K DO
  BEGIN
    X1 := new(PInfInt,Prod(X,Z));
    Z^.ChangeSign;
    Y1 := new(PInfInt,Prod(Z,Y));
    Z1 := new(PInfInt,Prod(X,X));
    H1 := new(PInfInt,Prod(Z1,N));
    dispose(Z1,Done);
    Z1 := new(PInfInt,Prod(Y,Y));
    dispose(X,Done);
    dispose(Y,Done);
    dispose(Z,Done);
    X := X1;
    Y := Y1;
    Z := new(PInfInt,Diff(H1,Z1));
    dispose(H1,Done);
    dispose(Z1,Done);
  END; {WITH}
END;{Kehrwert}

FUNCTION Kettenbruchentwicklung(N1 :PInfInt;VAR Periodenlaenge: integer;Ausg:boolean):Kettenbruch;
{Berechnet die erste Periode des Kettenbruchs der Wurzel von N1,falls Ausg true ist wird der Ketten-}
{bruch auch noch auf dem Bildschirm ausgegeben. }
VAR  Rest1, Rest : Koerperelement;
     Erg,ErgP,H  : Kettenbruch;
     A,B,C       : PInfInt;
BEGIN
  A := new(PInfInt,Assign(N1)); A^.Sqrt;
  B := new(PInfInt,Prod(A,A));
  IF B^.Compare(N1) >= 0 THEN
  BEGIN
    write(#7,'Fehler ');N1^.Write(Output) ;writeln(' ist Quadratzahl!');
    write('Wurzel = ');A^.write(Output); writeln;
    KettenBruchEntwicklung := nil;
    dispose(A,Done);
    dispose(B,Done);
    exit;
  END;
  new(Erg); Erg^.Ziffer := A;
  dispose(B,Done);
  IF Ausg THEN
  BEGIN
    write('[');A^.Write(Output);
  END;
  Erg^.next := nil;
  Rest.X := new(PInfInt,AssignLong(1));
  Rest.Y := new(PInfInt,Assign(Erg^.Ziffer));
  Rest.Y^.ChangeSign;
  Rest.Z := new(PInfInt,Assign(Rest.X));
  Rest1.X := new(PInfInt,Assign(Rest.X));
  Rest1.Y := new(PInfInt,Assign(Rest.Y));
  Rest1.Z := new(PInfInt,Assign(Rest.Z));
  Rest.N := N1;
  ErgP := Erg;
  Periodenlaenge := 0;
  REPEAT {Nchste Ziffer des Kettenbruchs suchen}
    kehrwert(Rest);
    A := new(PInfInt,Prod(Rest.X,Rest.X));
    A^.Multiply(N1);
    A^.Sqrt;
    A^.Add(Rest.Y);
    dispose(A^.DivideRemainder(Rest.Z),Done);
    new(ErgP^.next); ErgP := ErgP^.next;
    ErgP^.Next := nil; new(ErgP^.Ziffer,Assign(A));
    IF Ausg THEN
    BEGIN
      write(',');A^.write(Output);
    END;
    {neuen Rest berechnen}
    B := new(PInfInt,Prod(A,Rest.Z));
    dispose(A,Done);
    A := new(PInfInt,Diff(Rest.Y,B));
    dispose(B,Done);
    dispose(Rest.Y,Done); Rest.Y := A;
    kuerzen(Rest);
    Periodenlaenge := succ(Periodenlaenge);
  UNTIL (Rest.X^.Compare(Rest1.X) = 0) AND (Rest.Y^.Compare(Rest1.Y) = 0)
        AND (Rest.Z^.Compare(Rest1.Z) = 0);
  WITH Rest DO
  BEGIN
    dispose(X,Done);
    dispose(Y,Done);
    dispose(Z,Done);
  END;
  WITH Rest1 DO {Speicher wieder freigeben}
  BEGIN
    dispose(X,Done);
    dispose(Y,Done);
    dispose(Z,Done);
  END;
  IF Ausg THEN writeln(',...]');
  Kettenbruchentwicklung := erg;
END; {Kettenbruchentwicklung}

PROCEDURE Naeherungsbruch(K:Kettenbruch;n: integer;VAR Zaehler,Nenner : PInfInt;Ausg:boolean);
{Berechnet den n-ten Nherungsbruch des Kettenbruchs K. K wird als periodisch
 angenommen mit Periodenanfang im zweiten Element, wie dies bei Quadratwurzeln
 von ganzen Zahlen immer der Fall ist.Falls Ausg true ist werden alle Nherungs-
 brche auf dem Bildschirm ausgegeben.}
VAR pk,pk1,qk,qk1,h1,h2 : PInfInt;
    l                   : integer;
    Kptr                : Kettenbruch;
BEGIN
  pk := new(PInfInt,Assign(K^.Ziffer));
  pk1 := new(PInfInt,AssignLong(1));
  qk := new(PInfInt,AssignLong(1));
  qk1 := new(PInfInt,AssignLong(0));
  IF Ausg THEN
  BEGIN
    write('first rational approximation : ');pk^.Write(Output);writeln('/');
    write('                               '); qk^.Write(Output);
    readln;
  END;
  Kptr := K^.next;
  FOR l := 2 TO n DO
  BEGIN
    h1 := new(PInfInt,Prod(KPtr^.Ziffer,pk));
    h2 := new(PInfInt,Sum(h1,pk1));
    dispose(h1,Done);
    pk1^.Assignnew(pk);
    pk^.Assignnew(h2);
    dispose(h2,Done);
    h1 := new(PInfInt,Prod(Kptr^.Ziffer,qk));
    h2 := new(PInfInt,Sum(h1,qk1));
    dispose(h1,Done);
    qk1^.Assignnew(qk);
    qk^.Assignnew(h2);
    dispose(h2,Done);
    IF Kptr^.next <> nil THEN Kptr := Kptr^.next
    ELSE Kptr := K^.next;
    IF Ausg THEN
    BEGIN
      write(l,'-th rational approximation : ');pk^.Write(Output);writeln('/');
      write('                               '); qk^.Write(Output);
      readln;
    END;
  END;
  Zaehler := pk;
  Nenner := qk;
  dispose(pk1,Done);
  dispose(qk1,Done);
END;{Naeherungsbruch}

PROCEDURE Begruessung;
BEGIN
  {ClrScr; }
  writeln('                 PELL Vers 2.0  by C.Sperr 18.07.1993');
{  writeln('Dieses Programm berechnet zu einer Zahl N die Kettenbruchentwicklung der ');
  writeln('Quadratwurzel von N und dann wahlweise die k ersten Nherungsbrche dieser');
  writeln('Wurzel, oder eine Lsung der Pell''schen Gleichung X*X - N*Y*Y =  1. Dabei');
  writeln('gibt es fr das positive Vorzeichen immer eine Lsung, nicht aber fr das');
  writeln('negative. Neben X und Y wird auch noch X*X und N*Y*Y ausgegeben.');
  writeln('Es sind ',MemAvail,' bytes Heap-Speicher vorhanden');  }
  writeln('This program calculates the continued fraction of the squareroot of a');
  writeln('number n. After that it computes rational approximations to sqrt(n) or a');
  writeln('solution of the Pell''equation X*X - N*Y*Y =  1. ');
END;{Begruessung}

BEGIN {Pell}
  getmem(s,256);
  Begruessung; entwickelt := false;
  REPEAT
    write('N = ');
    readln(PcharArr(S)^);
    N := new(PInfInt,AssignStr(s));
  UNTIL N <> nil;
  REPEAT
    write('N)ew N, R)ational approximations, P)ells equation, E)nd: ');
    REPEAT readln(ch); ch := upcase(ch) UNTIL ch IN ['E','N','P','R'];
    writeln(ch);
    CASE ch OF
      'N' : BEGIN
              IF entwickelt THEN
              BEGIN
                Kptr := KB;
                WHILE Kptr <> nil DO
                BEGIN
                  KB := Kptr; Kptr := Kptr^.next;
                  dispose(KB^.Ziffer,Done);
                  dispose(KB);
                END;
              END;
              dispose(N,Done);entwickelt := false;
              REPEAT
                write('N = ');
                readln(PcharArr(S)^);
                N := new(PInfInt,AssignStr(s));
              UNTIL N <> nil;
            END;
      'R' : BEGIN
              IF NOT entwickelt THEN
              BEGIN
                write('Continued fraction of (');N^.Write(Output);writeln(')^');
                KB := Kettenbruchentwicklung(N,pl,true);
                IF KB <> nil THEN
                BEGIN
                  writeln('length of period of continued fraction : ',pl);
                  entwickelt := true;
                END
                ELSE
                  continue;
              END;
              write('Approximations up to index ');readln(k);
              Naeherungsbruch(KB,k,X,Y,true);
              dispose(x,Done); dispose(Y,Done);
            END;
      'P' : BEGIN
              IF NOT entwickelt THEN
              BEGIN
                write('Continued fraction of (');N^.write(Output);writeln(')^');
                KB := Kettenbruchentwicklung(N,pl,true);
                IF KB <> nil THEN
                BEGIN
                  writeln('Length of period of continued fraction : ',pl);
                  entwickelt := true;
                END
                ELSE
                  continue;
              END;
              IF odd(pl) THEN
              BEGIN
                Naeherungsbruch(KB,pl,X,Y,false);
                write('Solution of X*X - ');N^.write(Output);writeln('*Y*Y = -1');
                write('X = '); X^.Write(Output); writeln;
                write('Y = '); Y^.Write(Output); writeln;
                XX := new(PInfInt,Prod(X,X));
                write('  X*X  = ');XX^.Write(Output);writeln;
                dispose(XX,Done); XX := new(PInfInt,Prod(Y,Y));
                YY := new(PInfInt,Prod(XX,N)); dispose(XX,Done);
                write('N*Y*Y  = ');YY^.Write(Output);writeln;
                k := 2*pl; dispose(YY,Done); dispose(X,Done); dispose(Y,Done);
              END
              ELSE k := pl;
              Naeherungsbruch(KB,k,X,Y,false);
              write('Solution of X*X - ');N^.Write(OutPut);writeln('*Y*Y = 1');
              write('X = '); X^.Write(Output); writeln;
              write('Y = '); Y^.Write(Output); writeln;
              XX := new(PInfInt,Prod(X,X));
              write('  X*X  = ');XX^.Write(Output);writeln;
              dispose(XX,Done); XX := new(PInfInt,Prod(Y,Y));
              YY := new(PInfInt,Prod(XX,N)); dispose(XX,Done);
              write('N*Y*Y  = ');YY^.Write(Output);writeln;
              dispose(YY,Done); dispose(X,Done); dispose(Y,Done);
            END;
    END; {CASE}
  UNTIL ch = 'E';
END.{Pell}
